"""
All rule-based checks of calls to `df_to_figure` and `df_to_latex` are defined here.

The checks are divided into categories:

For the analysis step:
1. Syntax checks: checks for correct syntax of the calls to `df_to_figure` and `df_to_latex`.
2. Content checks (analysis): Check for content of the dataframes created in the analysis step.

For the display-item step, we further add:
3. Continuity checks: Check that the df are created from the output of the analysis step.
4. Content checks (display-item): Check for content of the dataframes created in the display-item step.
5. Compilation checks: Run compilation and check for errors and wide tables.
6. Annotation checks: Check for annotations (like the labels of the df, and note, glossary, and caption).


Note: check methods must start with `check_`.
Use CHOICE_OF_CHECKS to specify the order and which checks to run.

For transparency, all check_ methods must be specified in CHOICE_OF_CHECKS.
This is automatically asserted by the BaseChecker class.
"""


from dataclasses import dataclass, field
import re
from typing import Dict, Union, Optional, List, Any, Tuple, Type, ClassVar, Callable

import numpy as np
import pandas as pd
from pathlib import Path

from data_to_paper.env import DEBUG_MODE
from data_to_paper.latex.exceptions import BaseLatexProblemInCompilation
from data_to_paper.latex.latex_doc import LatexDocument
from data_to_paper.llm_coding_utils import df_to_latex, df_to_figure, ALLOWED_PLOT_KINDS, DF_ALLOWED_VALUE_TYPES
from data_to_paper.llm_coding_utils.consts import DF_ALLOWED_COLUMN_TYPES
from data_to_paper.llm_coding_utils.df_to_figure import run_create_fig_for_df_to_figure_and_get_axis_parameters
from data_to_paper.llm_coding_utils.matplotlib_utils import AxisParameters
from data_to_paper.run_gpt_code.base_run_contexts import RegisteredRunContext
from data_to_paper.run_gpt_code.overrides.dataframes.df_with_attrs import InfoDataFrameWithSaveObjFuncCall

from data_to_paper.text import dedent_triple_quote_str
from data_to_paper.utils.check_type import raise_on_wrong_func_argument_types, name_of_type
from data_to_paper.utils.dataframe import extract_df_row_labels, extract_df_column_labels, extract_df_axes_labels
from data_to_paper.utils.nice_list import NiceList
from data_to_paper.utils.numerics import is_lower_eq

from data_to_paper.research_types.hypothesis_testing.env import get_max_rows_and_columns, MAX_BARS
from data_to_paper.run_gpt_code.overrides.dataframes.utils import df_to_llm_readable_csv
from data_to_paper.run_gpt_code.overrides.pvalue import is_p_value, PValue, is_containing_p_value, is_only_p_values, \
    OnStrPValue, OnStr
from data_to_paper.run_gpt_code.run_issues import CodeProblem, RunIssue, RunIssues
from .abbreviations import is_unknown_abbreviation
from .utils import is_non_integer_numeric, _find_longest_labels_in_index, \
    _find_longest_labels_in_columns_relative_to_content


@dataclass
class BaseChecker:

    issues: RunIssues = field(default_factory=RunIssues)
    intermediate_results: Dict[str, Any] = field(default_factory=dict)
    stop_after_first_issue: bool = False

    def _append_issue(self, category: str = None, item: str = None, issue: str = '', instructions: str = '',
                      code_problem: CodeProblem = None, forgive_after: Optional[int] = None):
        self.issues.append(RunIssue(
            category=category,
            item=item,
            issue=issue,
            instructions=instructions,
            code_problem=code_problem,
            forgive_after=forgive_after,
        ))

    def _automatically_get_all_check_methods(self):
        return [getattr(type(self), method_name) for method_name in dir(self) if method_name.startswith('check')]

    def _assert_if_provided_choice_of_checks_include_all_check_methods(self):
        if self.CHOICE_OF_CHECKS is not None:
            for check_method in self._automatically_get_all_check_methods():
                assert check_method in self.CHOICE_OF_CHECKS, f'Missing check method: {check_method.__name__}'

    def _get_checks_to_run(self):
        if self.CHOICE_OF_CHECKS is None:
            return self._automatically_get_all_check_methods()
        return [check for check, should_check in self.CHOICE_OF_CHECKS.items() if should_check]

    def _run_checks(self):
        self._assert_if_provided_choice_of_checks_include_all_check_methods()
        for check in self._get_checks_to_run():
            num_issues_before = len(self.issues)
            should_stop = check(self)
            num_created_issues = len(self.issues) - num_issues_before
            if should_stop and not num_created_issues:
                assert False, f'Check {check.__name__} returned True, but no issues were created.'
            if DEBUG_MODE:
                print(f'Check "{check.__name__}" created {num_created_issues} issues.')
            if self.issues and (self.stop_after_first_issue or should_stop):
                break

    def run_checks(self) -> Tuple[RunIssues, Dict[str, Any]]:
        self._run_checks()
        return self.issues, self.intermediate_results

    CHOICE_OF_CHECKS: ClassVar[Optional[Dict[Callable, bool]]] = None


@dataclass
class ChainChecker(BaseChecker):
    checkers: List[BaseChecker] = None
    stop_after_first_issue: bool = True

    def _run_checks(self):
        for checker in self.checkers:
            checker.intermediate_results.update(self.intermediate_results)
            issues, intermediate_results = checker.run_checks()
            self.issues.extend(issues)
            self.intermediate_results.update(intermediate_results)
            non_forgivable_issues = [issue for issue in self.issues if issue.forgive_after is None]
            if non_forgivable_issues and self.stop_after_first_issue:
                break


def create_and_run_chain_checker(checkers: List[Type[BaseChecker]], stop_after_first_issue: bool = True, **kwargs
                                 ) -> Tuple[RunIssues, Dict[str, Any]]:
    chain_checker = ChainChecker(checkers=[checker(**kwargs) for checker in checkers],  # type: ignore
                                 stop_after_first_issue=stop_after_first_issue)
    return chain_checker.run_checks()


@dataclass
class BaseDfChecker(BaseChecker):
    func: Callable = None
    df: pd.DataFrame = None
    filename: str = None
    kwargs: dict = field(default_factory=dict)

    output_folder: Optional[Path] = None  # where compiled figures and tables are saved
    latex_document: Optional[LatexDocument] = None

    DEFAULT_CATEGORY: ClassVar[str] = None
    DEFAULT_CODE_PROBLEM: ClassVar[CodeProblem] = None

    NAMES_OF_KWARGS_FOR_LATEX = ['caption', 'label', 'note', 'glossary']

    def _append_issue(self, category: str = None, item: str = None, issue: str = '', instructions: str = '',
                      code_problem: CodeProblem = None, forgive_after: Optional[int] = None):
        category = self.DEFAULT_CATEGORY if category is None else category
        code_problem = self.DEFAULT_CODE_PROBLEM if code_problem is None else code_problem
        item = self.filename if item is None else item
        super()._append_issue(category=category, item=item, issue=issue, instructions=instructions,
                              code_problem=code_problem, forgive_after=forgive_after)

    def _run_and_get_result_and_exception(self, **additional_kwargs) -> Tuple[Any, Optional[Exception]]:
        try:
            result = self.func(self.df, self.filename, **self.kwargs, **additional_kwargs)
        except Exception as e:
            return None, e
        return result, None

    @property
    def func_name(self):
        if self.func is None:
            return None
        return self.func.__name__

    @property
    def is_figure(self):
        return self.func_name == 'df_to_figure'

    @property
    def table_or_figure(self):
        return 'figure' if self.is_figure else 'table'

    @property
    def index(self) -> bool:
        if self.is_figure:
            return self.kwargs.get('use_index', True) and self.x is None
        else:
            return self.kwargs.get('index', True)

    def get_x_labels(self):
        if self.is_figure:
            if self.index:
                return self.df.index
            x = self.x
            if x is None:
                return np.array(range(self.df.shape[0]))
            return self.df[x]
        else:
            if self.index:
                return self.df.index
            # if the index is not used, it is the first column that behaves as the index:
            return self.df.iloc[:, 0]

    def get_y_labels(self):
        if self.is_figure:
            y, _, _, _ = self.get_xy_err_ci_p_value('y', as_list=True)
            return y
        return self.df.columns

    @property
    def note(self) -> Optional[str]:
        return self.kwargs.get('note', None)

    @property
    def glossary(self) -> Optional[Dict[str, str]]:
        return self.kwargs.get('glossary', None)

    @property
    def caption(self) -> Optional[str]:
        return self.kwargs.get('caption', None)

    @property
    def kind(self) -> Optional[str]:
        return self.kwargs.get('kind', 'bar')

    @property
    def x(self) -> Optional[str]:
        return self.kwargs.get('x', None)

    @property
    def y(self) -> Optional[Union[str, List[str]]]:
        return self.kwargs.get('y', None)

    @property
    def yerr(self) -> Optional[Union[str, List[str]]]:
        return self.kwargs.get('yerr', None)

    @property
    def xerr(self) -> Optional[str]:
        return self.kwargs.get('yerr', None)

    @property
    def y_ci(self) -> Optional[Union[str, List[str]]]:
        return self.kwargs.get('y_ci', None)

    @property
    def x_ci(self) -> Optional[str]:
        return self.kwargs.get('x_ci', None)

    @property
    def y_p_value(self) -> Optional[Union[str, List[str]]]:
        return self.kwargs.get('y_p_value', None)

    @property
    def x_p_value(self) -> Optional[str]:
        return self.kwargs.get('x_p_value', None)

    def get_xlabel(self):
        label = self.kwargs.get('xlabel', None)
        if label is not None:
            return label
        return self.x

    def get_ylabel(self):
        label = self.kwargs.get('ylabel', None)
        if label is not None:
            return label
        if not isinstance(self.y, list):
            return self.y
        return None

    @staticmethod
    def _convert_to_list(value, flatten):
        if value is None:
            as_list = []
        else:
            as_list = [value] if isinstance(value, str) else value

        if not flatten:
            return as_list

        flatten_list = []
        for item in as_list:
            if isinstance(item, tuple):
                flatten_list.extend(item)
            else:
                flatten_list.append(item)
        return flatten_list

    def get_xy_err_ci_p_value(self, x_or_y: str, as_list=False, flatten=False):

        xy = getattr(self, x_or_y)
        err = getattr(self, f'{x_or_y}err')
        ci = getattr(self, f'{x_or_y}_ci')
        p_value = getattr(self, f'{x_or_y}_p_value')
        if as_list:
            xy = self._convert_to_list(xy, flatten)
            err = self._convert_to_list(err, flatten)
            ci = self._convert_to_list(ci, flatten)
            p_value = self._convert_to_list(p_value, flatten)
        return xy, err, ci, p_value

    @property
    def kwargs_for_plot(self):
        """
        Remove the kwargs for latex from the kwargs.
        """
        return {key: value for key, value in self.kwargs.items() if key not in self.NAMES_OF_KWARGS_FOR_LATEX}

    CHOICE_OF_CHECKS = {}


""" SYNTAX """


@dataclass
class SyntaxDfChecker(BaseDfChecker):
    """
    Checks that do not depend on the content of df.
    """
    DEFAULT_CATEGORY = 'Checking df_to_figure/df_to_latex for call syntax'
    DEFAULT_CODE_PROBLEM = CodeProblem.OutputFileCallingSyntax

    def check_argument_types(self):
        try:
            raise_on_wrong_func_argument_types(df_to_latex, self.df, self.filename, **self.kwargs)
        except TypeError as e:
            self._append_issue(issue=str(e))
            return True  # No more checks. Further checks may create errors.

    def check_filename(self):
        """
        Check if the filename of in the format `df_<alphanumeric>`.
        """
        pattern = r'^df_[a-zA-Z0-9_]+$'
        if not re.match(pattern=pattern, string=self.filename):
            self._append_issue(
                issue=f'Filenames of tables and figures should be in the format `df_<alphanumeric>` (without .ext), '
                      f'but got "{self.filename}".',
            )

    def check_no_label(self):
        if self.kwargs.get('label', None):
            self._append_issue(
                issue=f'The `label` argument should not be used in `df_to_figure` or `df_to_latex`; '
                      f'It is automatically generated from the filename.',
                instructions='Please remove the `label` argument.',
            )

    CHOICE_OF_CHECKS = BaseDfChecker.CHOICE_OF_CHECKS | {
        check_argument_types: True,
        check_filename: True,
        check_no_label: True,
    }


@dataclass
class TableSyntaxDfChecker(SyntaxDfChecker):
    func: Callable = df_to_latex

    def check_that_index_is_true(self):
        if not self.index:
            self._append_issue(
                issue='Do not call `df_to_latex` with `index=False`.',
                instructions=dedent_triple_quote_str("""
                    Please revise the code making sure all tables are created with `index=True`, \t
                    and that the index is meaningful.
                    """)
            )

    def check_column_arg_is_not_used(self):
        if 'columns' in self.kwargs:
            self._append_issue(
                issue='Do not use the `columns` argument in `df_to_latex`.',
                instructions='If you want to drop columns, do it before calling `df_to_latex`.',
            )

    CHOICE_OF_CHECKS = SyntaxDfChecker.CHOICE_OF_CHECKS | {
        check_that_index_is_true: True,
        check_column_arg_is_not_used: True,
    }


@dataclass
class FigureSyntaxDfChecker(SyntaxDfChecker):
    func: Callable = df_to_figure

    def check_kind_arg(self):
        """
        Check if the plot kind is one of the allowed plot kinds.
        """
        if 'kind' not in self.kwargs:
            self._append_issue(
                issue=f'Plot `kind` is not specified.',
                instructions=f'Please explicitly specify the `kind` argument. available options are:\n'
                             f'{ALLOWED_PLOT_KINDS}.',
            )
        elif self.kind not in ALLOWED_PLOT_KINDS:
            self._append_issue(
                issue=f'Plot kind "{self.kind}" is not supported.',
                instructions=f'Only use these kinds: {ALLOWED_PLOT_KINDS}.',
            )

    def check_y_arg(self):
        if self.y is None:
            self._append_issue(
                issue=f'No y values are specified.',
                instructions='Please use the `y` argument to specify the columns to be plotted.',

            )

    def check_yerr_arg(self):
        if self.yerr is not None:
            self._append_issue(
                issue=f'Do not use the `yerr` argument in `df_to_figure`.',
                instructions='Instead, directly indicate the confidence intervals using the `y_ci` argument.',
            )

    def check_x_p_value_arg(self):
        if self.x_p_value:
            self._append_issue(
                issue='The `x_p_value` argument is not supported.',
                instructions='Please use the `y_p_value` argument instead.',
            )

    def check_y_p_value_arg(self):
        # This check i obsolete, because of `check_x_p_value_arg`.
        if self.x_p_value is not None and self.y_p_value is not None:
            self._append_issue(
                issue='Both `x_p_value` and `y_p_value` are set.',
                instructions='Please use only one of them.',
            )
            return

    CHOICE_OF_CHECKS = SyntaxDfChecker.CHOICE_OF_CHECKS | {
        check_kind_arg: True,
        check_y_arg: True,
        check_yerr_arg: True,
        check_x_p_value_arg: True,
        check_y_p_value_arg: True,
    }


""" CONTENT FOR ANALYSIS STEP """


@dataclass
class BaseContentDfChecker(BaseDfChecker):
    stop_after_first_issue: bool = True

    prior_dfs: Dict[str, pd.DataFrame] = field(default_factory=dict)

    ALLOWED_VALUE_TYPES = DF_ALLOWED_VALUE_TYPES + (PValue,)
    ALLOWED_COLUMN_AND_INDEX_TYPES = {'columns': DF_ALLOWED_COLUMN_TYPES, 'index': DF_ALLOWED_VALUE_TYPES}
    ALLOW_MULTI_INDEX_FOR_COLUMN_AND_INDEX = {'columns': True, 'index': True}

    DEFAULT_CATEGORY = 'Checking content of created dfs'
    DEFAULT_CODE_PROBLEM = CodeProblem.OutputFileContentA

    def _get_x_values(self):
        if self.is_figure:
            x = self.x
            if x is None:
                return self.df.index
            return self.df[x]
        else:
            if self.index:
                return self.df.index
            # if the index is not used, it is the first column that behaves as the index:
            return self.df.iloc[:, 0]

    def _get_max_rows_and_columns(self):
        return get_max_rows_and_columns(self.is_figure, kind=self.kind, to_show=False)


@dataclass
class DfContentChecker(BaseContentDfChecker):
    VALUES_CATEGORY = 'Problem with df values'
    INDEX_COLUMN_CATEGORY = 'Problem with df index/columns'
    SIZE_CATEGORY = 'Too large df'

    def get_relevant_part_of_df(self):
        """
        Get the part of the df that is relevant for the checks.
        """
        if not self.is_figure:
            return self.df
        y, yerr, y_ci, y_p_value = self.get_xy_err_ci_p_value('y', as_list=True, flatten=True)
        x, xerr, x_ci, x_p_value = self.get_xy_err_ci_p_value('x', as_list=True, flatten=True)
        return self.df[y + yerr + y_ci + y_p_value + x + xerr + x_ci + x_p_value]

    def check_df_headers_type(self):
        for column_or_index in ['columns', 'index']:
            headers = getattr(self.df, column_or_index)

            # Check if the headers are a multi-index and if it is allowed:
            if not self.ALLOW_MULTI_INDEX_FOR_COLUMN_AND_INDEX[column_or_index] and isinstance(headers, pd.MultiIndex):
                self._append_issue(
                    category=self.INDEX_COLUMN_CATEGORY,
                    issue=f"Your dataframe has a multi-index for the {column_or_index}.",
                    instructions=f"Please make sure the df has a single-level {column_or_index}.",
                )
                continue

            # Check if the headers are of the allowed types:
            allowed_types = self.ALLOWED_COLUMN_AND_INDEX_TYPES[column_or_index]
            unsupported_headers = self._get_df_headers_of_unsupported_types(headers, allowed_types)
            if unsupported_headers:
                nice_list = NiceList((f"`{repr(header)}`  (type `{name_of_type(type(header))}` is not allowed)"
                                      for header in unsupported_headers), separator='\n')
                nice_allowed_types = [name_of_type(allowed_type) for allowed_type in allowed_types]
                self._append_issue(
                    category=self.INDEX_COLUMN_CATEGORY,
                    issue=f"Your df has {column_or_index} headers of unsupported types:\n{nice_list}.",
                    instructions=f"The df {column_or_index} headers should be only of these types: "
                                 f"{nice_allowed_types}.",
                )

    def _check_if_df_within_df(self) -> bool:
        for value in self.df.values.flatten():
            if isinstance(value, (pd.Series, pd.DataFrame)):
                self._append_issue(
                    category=self.VALUES_CATEGORY,
                    issue=f"Something wierd in your dataframe. Iterating over df.values.flatten() "
                          f"returned a `{type(value).__name__}` object.",
                )
                return True
        return False

    def check_df_value_types(self):
        """
        Check if the dataframe has only allowed value types.
        """
        if self._check_if_df_within_df():
            return
        un_allowed_type_names = {f'{type(value).__name__}' for value in self.df.values.flatten()
                                 if not isinstance(value, self.ALLOWED_VALUE_TYPES)}
        if un_allowed_type_names:
            self._append_issue(
                category=self.VALUES_CATEGORY,
                issue=f"Your dataframe contains values of types {sorted(un_allowed_type_names)} which are not allowed.",
                instructions=f"Please make sure the saved dataframes have only numeric, str, bool, or tuple values.",
            )

    def check_df_for_nan_values(self):
        """
        Check if the df has NaN values or PValue with value of nan
        """
        relevant_df = self.get_relevant_part_of_df()
        df_with_raw_pvalues = relevant_df.applymap(lambda v: v.value if is_p_value(v) else v)
        isnull = pd.isnull(df_with_raw_pvalues)
        num_nulls = isnull.sum().sum()
        if num_nulls > 0:
            issue_text = f'Note that the df has {num_nulls} NaN value(s).'
            if len(isnull) < 20:
                issue_text += f'\nHere is the `isnull` of the df:'
            else:
                # show only the first 10 rows with NaN values:
                isnull = self.df[isnull.any(axis=1)].head(10)
                issue_text += f'\nHere are some example lines with NaN values:'
            issue_text += f'\n```\n{df_to_llm_readable_csv(isnull)}\n```\n'
            instructions = \
                f"Please revise the code to avoid NaN values in the created {self.table_or_figure}."
            if not self.is_figure:
                instructions += \
                    "\nIf the NaNs are legit and stand for missing values: replace them with the string '-'.\n" \
                    "Otherwise, if they are computational errors, please revise the code to fix it."
            self._append_issue(
                category=self.VALUES_CATEGORY,
                issue=issue_text,
                instructions=instructions
            )

    @staticmethod
    def _get_df_headers_of_unsupported_types(headers: Union[pd.MultiIndex, pd.Index], allowed_types: Tuple[Type]
                                             ) -> List[Any]:
        """
        Find any headers of the dataframe are int, str, or bool.
        """
        if isinstance(headers, pd.MultiIndex):
            headers = [label for level in range(headers.nlevels) for label in headers.get_level_values(level)]
        return [header for header in headers if not isinstance(header, allowed_types)]

    @staticmethod
    def _is_index_a_range(index, max_allowed_range: int = 2):
        num_rows = len(index)
        return not isinstance(index, pd.MultiIndex) \
            and list(index) == list(range(num_rows)) and pd.api.types.is_integer_dtype(index.dtype) \
            and num_rows > max_allowed_range

    CHOICE_OF_CHECKS = BaseDfChecker.CHOICE_OF_CHECKS | {
        check_df_headers_type: True,
        check_df_for_nan_values: True,
        check_df_value_types: True,
    }


@dataclass
class TableDfContentChecker(DfContentChecker):
    func: Callable = df_to_latex

    OVERLAYING_VALUES_CATEGORY = 'Overlapping values'
    DF_DISPLAY_CATEGORY = 'The df looks like a df.describe() table, not a scientific table'

    MAX_ROWS_ALLOWED_FOR_RANGE_AS_INDEX = 2

    def check_df_is_a_result_of_describe(self):
        """
        Check if the table is a df.describe() table
        """
        description_labels = ('mean', 'std', 'min', '25%', '50%', '75%', 'max')
        if set(description_labels).issubset(self.df.columns) or set(description_labels).issubset(self.df.index):
            self._append_issue(
                category=self.DF_DISPLAY_CATEGORY,
                issue=f'The df includes mean, std, as well as quantiles and min/max values.',
                instructions=dedent_triple_quote_str("""
                    Note that in scientific tables, it is not customary to include quantiles, or min/max values, \t
                    especially if the mean and std are also included.
                    Please revise the code so that the tables only include scientifically relevant statistics.
                    """),
                forgive_after=3,
            )

    def check_df_index_is_a_range(self):
        """
        Check if the index of the dataframe is just a numeric range.
        """
        if not self.index:
            return
        num_rows = self.df.shape[0]
        if self._is_index_a_range(self.df.index, self.MAX_ROWS_ALLOWED_FOR_RANGE_AS_INDEX):
            self._append_issue(
                category=self.INDEX_COLUMN_CATEGORY,
                issue=dedent_triple_quote_str(f"""
                    The function `df_to_latex` uses the index as the row labels.
                    But, the index of df "{self.filename}" is just a range from 0 to {num_rows - 1}.
                    """),
                instructions=dedent_triple_quote_str(f"""
                    Please revise the code making sure the figure is built with an index that represents meaningful \t
                    numeric data. Or, for categorical data, the index should be a list of strings.

                    Note: labeling row with sequential numbers is not common in scientific tables. \t
                    But, if you are sure that starting each row with a sequential number is really what you want, \t
                    then convert it from int to strings, so that it is clear that it is not a mistake.
                    """),
            )

    def check_df_for_repeated_values(self):
        """
        # Check if the table contains the same values in multiple cells
        """
        df_values = [v for v in self.df.values.flatten() if is_non_integer_numeric(v)]
        if len(df_values) != len(set(df_values)):
            # Find the positions of the duplicated values:
            duplicated_values = [v for v in df_values if df_values.count(v) > 1]
            example_value = duplicated_values[0]
            duplicated_value_positions = np.where(self.df.values == example_value)
            duplicated_value_positions = list(zip(*duplicated_value_positions))
            duplicated_value_positions = [f'({row}, {col})' for row, col in duplicated_value_positions]
            duplicated_value_positions = ', '.join(duplicated_value_positions)

            self._append_issue(
                category=self.OVERLAYING_VALUES_CATEGORY,
                issue=f'Note that the df "{self.filename}" includes the same values in multiple cells.\n'
                      f'For example, the value {example_value} appears in the following cells:\n'
                      f'{duplicated_value_positions}.',
                instructions=dedent_triple_quote_str("""
                    Is this perhaps a mistake, please revise the code so that the df does not repeat the same values \t
                    in multiple cells.
                    Otherwise, please just add a comment in the codee explaining why the same value might be repeated.
                    """),
                forgive_after=1,
            )

    def check_df_for_repeated_values_in_prior_dfs(self):
        """
        Check if the df numeric values overlap with values in prior dfs
        """
        if not self.prior_dfs:
            return
        df_values = [v for v in self.df.values.flatten() if is_non_integer_numeric(v)]
        for prior_name, prior_table in self.prior_dfs.items():
            if prior_table is self.df:
                continue
            prior_table_values = [v for v in prior_table.values.flatten() if is_non_integer_numeric(v)]
            if any(value in prior_table_values for value in df_values):
                self._append_issue(
                    category=self.OVERLAYING_VALUES_CATEGORY,
                    issue=f'Table "{self.filename}" includes values that overlap with values in table "{prior_name}".',
                    instructions=dedent_triple_quote_str("""
                        In scientific tables, it is not customary to include the same values in multiple tables.
                        Please revise the code so that each table include its own unique data.
                        """),
                    forgive_after=1,
                )

    def check_df_size(self):
        """
        Check if the df has too many columns or rows
        """
        shape = self.df.shape
        max_rows_and_columns = self._get_max_rows_and_columns()
        if is_lower_eq(shape[0], max_rows_and_columns[0]) and is_lower_eq(shape[1], max_rows_and_columns[1]):
            return
        max_rows, max_columns = max_rows_and_columns
        instructions = dedent_triple_quote_str(f"""
            Please revise the code so that df of created Tables \t
            have a maximum of {max_rows} rows and {max_columns} columns.
            Note that simply trimming the data is typically not a good solution.
            You might instead consider a different representation/organization of the table.
            Or, consider representing the data as a figure.
            """)
        if is_lower_eq(shape[0], max_rows_and_columns[1]) and is_lower_eq(shape[1], max_rows_and_columns[0]):
            instructions += "\nYou might also want to consider transposing the df (df = df.T)."
        for ax, rows_or_columns in enumerate(('rows', 'columns')):
            if not is_lower_eq(shape[ax], max_rows_and_columns[ax]):
                self._append_issue(
                    category=self.SIZE_CATEGORY,
                    issue=f'This table has {shape[ax]} {rows_or_columns}, which is too many for '
                          f'a scientific table (max allowed: {max_rows_and_columns[ax]}).',
                    instructions=instructions
                )

    CHOICE_OF_CHECKS = {
        check_df_is_a_result_of_describe: True,  # We want to start with detecting describe tables.
        check_df_index_is_a_range: True,
        check_df_for_repeated_values: True,
        check_df_for_repeated_values_in_prior_dfs: True,
        check_df_size: True,
    } | DfContentChecker.CHOICE_OF_CHECKS


@dataclass
class FigureDfContentChecker(DfContentChecker):
    func: Callable = df_to_figure
    ALLOW_MULTI_INDEX_FOR_COLUMN_AND_INDEX = {'columns': False, 'index': False}

    DEFAULT_CATEGORY = 'Checking figure'
    P_VALUE_CATEGORY = 'Plotting P-values'

    def check_that_specified_columns_exist(self):
        """
        Check if the columns specified in the `y` and `yerr` arguments exist in the df.
        """
        base_arg_names = ['', 'err', '_ci', '_p_value']
        for xy in ['x', 'y']:
            args = self.get_xy_err_ci_p_value(xy, as_list=True, flatten=True)
            arg_names = [xy + arg_name for arg_name in base_arg_names]
            for arg, arg_name in zip(args, arg_names):
                if arg is None:
                    continue
                un_specified_columns = [col for col in arg if col not in self.df.columns]
                if un_specified_columns:
                    self._append_issue(
                        issue=f'The columns {un_specified_columns} specified in the `{arg_name}` argument do not exist '
                              f'in the df.',
                        instructions=f'Available columns are: {self.df.columns}.',
                    )

    def check_df_size(self):
        """
        Check if the df has too many columns or rows
        """
        shape = self.get_relevant_part_of_df().shape
        max_rows_and_columns = self._get_max_rows_and_columns()
        if is_lower_eq(shape[0], max_rows_and_columns[0]) and is_lower_eq(shape[1], max_rows_and_columns[1]):
            return
        max_rows, max_columns = max_rows_and_columns
        if not is_lower_eq(shape[0], max_rows):
            self._append_issue(
                category=self.SIZE_CATEGORY,
                issue=f'The df of this figure has {shape[0]} rows, which is too many '
                      f'(max allowed: {max_rows}).',
                instructions=dedent_triple_quote_str(f"""
                        Please revise the code so that this figure df shows a maximum of {max_rows} rows.
                        Note that simply trimming the data is typically not a good solution.
                        Carefully consider the data that is most important to show, and remove other rows.
                        Or, consider representing the data more aggregated.
                """)
            )
        num_series = len(self.get_y_labels())
        columns_per_series = shape[1] / num_series
        max_series = int(max_columns / columns_per_series)
        if not is_lower_eq(shape[1], max_columns) and num_series > 2:
            self._append_issue(
                category=self.SIZE_CATEGORY,
                issue=f'This figure has {num_series} series which is too many for a single plot.',
                instructions=f'Please revise the code so that this figure df shows a maximum of {max_series} series.'
            )

    def check_that_y_values_are_numeric(self):
        y, yerr, y_ci, y_p_value = self.get_xy_err_ci_p_value('y', as_list=True)
        for column in y:
            if not pd.api.types.is_numeric_dtype(self.df[column]):
                self._append_issue(
                    issue=f'Column `{column}` is not numeric, so it is not suitable for a plot.',
                    instructions='All columns specified by the `y` argument must have numeric values.',
                )

    def check_that_x_values_are_numeric(self):
        x, xerr, x_ci, x_p_value = self.get_xy_err_ci_p_value('x', as_list=True)
        for column in x:
            if not pd.api.types.is_numeric_dtype(self.df[column]):
                self._append_issue(
                    issue=f'The x-column `{column}` is not numeric.',
                    instructions='If you are attempting to define non-numerical x labels, '
                                 'move the column to the index of the df, and set `x=None` in df_to_figure.\n'
                                 'Otherwise, only use the `x` argument for numeric x values.',
                )

    def _get_columns_containing_p_values(self):
        return [col for col in self.df.columns if is_containing_p_value(self.df[col])]

    def _get_columns_with_p_values_only(self):
        return [col for col in self.df.columns if is_only_p_values(self.df[col])]

    def check_for_mixing_p_values(self):
        # check that the columns with p-values only contain p-values:
        containing_p_values = self._get_columns_containing_p_values()
        not_pure_p_values = [col for col in containing_p_values if col not in self._get_columns_with_p_values_only()]
        if not_pure_p_values:
            self._append_issue(
                category=self.P_VALUE_CATEGORY,
                issue=f'The df has columns {not_pure_p_values}, which contain p-values and non-p-values.',
                instructions='Please make sure that the columns with p-values only contain p-values.',
            )

    def check_that_y_p_value_are_p_values(self):
        p_value_columns = self._get_columns_with_p_values_only()
        y, yerr, y_ci, y_p_value = self.get_xy_err_ci_p_value('y', as_list=True)
        chosen_columns_are_not_p_values = [col for col in y_p_value if col not in p_value_columns]
        if chosen_columns_are_not_p_values:
            self._append_issue(
                category=self.P_VALUE_CATEGORY,
                issue=f'The columns y_p_value={chosen_columns_are_not_p_values} are not p-values.',
                instructions='Please make sure that the columns with p-values only contain p-values.',
            )

    def check_all_p_values_are_in_y_p_value(self):
        p_value_columns = self._get_columns_with_p_values_only()
        y, yerr, y_ci, y_p_value = self.get_xy_err_ci_p_value('y', as_list=True)
        p_value_columns_not_in_y_p_value = [col for col in p_value_columns if col not in y_p_value]
        if p_value_columns_not_in_y_p_value:
            self._append_issue(
                category=self.P_VALUE_CATEGORY,
                issue=f'The columns {p_value_columns_not_in_y_p_value} contain p-values but are not in y_p_value.',
                instructions='Please include all the columns with p-values in y_p_value argument, or remove them '
                             'from the df before plotting.',
                forgive_after=1,
            )

    def check_for_max_number_of_bars(self):
        if self.kind not in ['bar', 'barh']:
            return
        y, yerr, y_ci, y_p_value = self.get_xy_err_ci_p_value('y', as_list=True)
        n_bars = len(self.df) * len(y)
        if not is_lower_eq(n_bars, MAX_BARS):
            self._append_issue(
                issue=f'The plot has {n_bars} bars, which is a large number.',
                instructions='Consider reducing the number of bars to make the plot more readable.',
                forgive_after=2,
            )

    def check_that_y_values_are_diverse(self):
        # There is no point in plotting a box, violin, hist if the y values are not diverse
        if self.kind not in ['box', 'violin', 'hist']:
            return
        y, yerr, y_ci, y_p_value = self.get_xy_err_ci_p_value('y', as_list=True)
        for column in y:
            n_unique = self.df[column].nunique()
            if n_unique <= 2:
                self._append_issue(
                    issue=f'Column `{column}` has only {n_unique} unique values, so it is not suitable for '
                          f'a "{self.kind}" plot.',
                    instructions='Choose another kind of plot, like calculating the mean and plotting a bar plot.',
                )

    def check_for_numeric_x_for_line_and_scatter(self):
        """check that we do not have non-numeric x:"""
        if self.kind not in ['line', 'scatter']:
            return
        if not pd.api.types.is_numeric_dtype(self._get_x_values()):
            self._append_issue(
                issue=f'The x values are not numeric, so they are not suitable for a "{self.kind}" plot.',
                instructions='Consider another kind of plot, like a bar plot (kind="bar").',
            )

    CHOICE_OF_CHECKS = {
        DfContentChecker.check_df_headers_type: True,
        check_that_specified_columns_exist: True,
        DfContentChecker.check_df_for_nan_values: True,
        DfContentChecker.check_df_value_types: True,
        check_for_max_number_of_bars: True,
        check_df_size: True,
        check_that_y_values_are_numeric: True,
        check_that_x_values_are_numeric: True,
        check_for_mixing_p_values: True,
        check_that_y_p_value_are_p_values: True,
        check_all_p_values_are_in_y_p_value: True,  # TODO: maybe disable this check
        check_that_y_values_are_diverse: True,
        check_for_numeric_x_for_line_and_scatter: True,
    }


""" COMPILATION """


@dataclass
class CompilationDfContentChecker(BaseContentDfChecker):
    intermediate_results: Dict[str, Any] = field(default_factory=lambda: {'width': None})
    tolerance_for_too_wide_in_pts: Optional[float] = 25.  # If None, do not raise on too wide.
    DEFAULT_CATEGORY = 'Table/Figure compilation failure'

    def check_creating_latex_and_html(self):
        for is_html in [False, True]:
            with OnStrPValue(OnStr.SMALLER_THAN):
                result, e = self._run_and_get_result_and_exception(is_html=is_html)
            if e is not None:
                self._append_issue(
                    issue=f'Failed to create the table. Got:\n{e}',
                    instructions='Please revise the code to fix the issue.',
                )
                return True

    CHOICE_OF_CHECKS = BaseContentDfChecker.CHOICE_OF_CHECKS | {
        check_creating_latex_and_html: True,
    }


@dataclass
class FigureCompilationDfContentChecker(CompilationDfContentChecker):
    func: Callable = df_to_figure

    def check_compilation_and_create_the_figure(self):
        """
        Axis parameters are returned and stored in intermediate_results.
        """
        if self.output_folder is None:
            return
        filepath = self.output_folder / f'{self.filename}.png'
        axis_parameters, exception = \
            run_create_fig_for_df_to_figure_and_get_axis_parameters(self.df, filepath, **self.kwargs_for_plot)
        if exception is not None:
            self._append_issue(
                issue=f'Failed to create the figure. Got:\n{exception}',
                instructions='Please revise the code to fix the issue.',
            )
            return True
        self.intermediate_results['axis_parameters'] = axis_parameters

    CHOICE_OF_CHECKS = CompilationDfContentChecker.CHOICE_OF_CHECKS | {
        check_compilation_and_create_the_figure: True,
    }


@dataclass
class TableCompilationDfContentChecker(CompilationDfContentChecker):
    func: Callable = df_to_latex

    def _df_to_latex_transpose(self):
        assert 'columns' not in self.kwargs, "assumes columns is None"
        kwargs = self.kwargs.copy()
        index = kwargs.pop('index', True)
        header = kwargs.pop('header', True)
        header, index = index, header
        return df_to_latex(self.df.T, self.filename, index=index, header=header, **kwargs)

    def check_compilation_and_get_width(self):
        exception = None
        with RegisteredRunContext.temporarily_disable_all():
            with OnStrPValue(OnStr.SMALLER_THAN):
                latex = df_to_latex(self.df, self.filename, **self.kwargs)
            try:
                width = self.latex_document.compile_table(latex)
            except BaseLatexProblemInCompilation as e:
                exception = e
                width = None

        # save the width of the table:
        self.intermediate_results['width'] = width

        if exception:
            self._append_issue(
                category='Table pdflatex compilation failure',
                issue=dedent_triple_quote_str("""
                    Here is the created table:

                    ```latex
                    {table}
                    ```

                    When trying to compile it using pdflatex, I got the following error:

                    {error}

                    """).format(filename=self.filename, table=latex, error=exception),
            )
        elif width > 1.3:
            # table is too wide
            # Try to compile the transposed table:
            with OnStrPValue(OnStr.SMALLER_THAN):
                latex_transpose = self._df_to_latex_transpose()
            e_transpose = None
            with RegisteredRunContext.temporarily_disable_all():
                try:
                    width_transpose = self.latex_document.compile_table(latex_transpose)
                except BaseLatexProblemInCompilation as e:
                    e_transpose = e
                    width_transpose = None
            if not e_transpose and width_transpose < 1.1:
                transpose_message = '- Alternatively, consider completely transposing the table. Use `df = df.T`.'
            else:
                transpose_message = ''
            index_note = ''
            column_note = ''
            if self.index:
                longest_index_labels = _find_longest_labels_in_index(self.df.index)
                longest_index_labels = [label for label in longest_index_labels if label is not None and len(label) > 6]
                with OnStrPValue(OnStr.SMALLER_THAN):
                    longest_column_labels = _find_longest_labels_in_columns_relative_to_content(self.df)
                longest_column_labels = [label for label in longest_column_labels if len(label) > 6]
                if longest_index_labels:
                    index_note = dedent_triple_quote_str(f"""\n
                        - Rename any long index labels to shorter names \t
                        (for instance, some long label(s) in the index are: {longest_index_labels}). \t
                        Use `df.rename(index=...)`
                        """)

                if longest_column_labels:
                    column_note = dedent_triple_quote_str(f"""\n
                        - Rename any long column labels to shorter names \t
                        (for instance, some long label(s) in the columns are: {longest_column_labels}). \t
                        Use `df.rename(columns=...)`
                        """)

            if not index_note and not column_note and not transpose_message:
                drop_column_message = dedent_triple_quote_str("""\n
                    - Drop unnecessary columns. \t
                    If the labels cannot be shortened much, consider whether there might be any \t
                    unnecessary columns that we can drop. \t
                    Use `df_to_latex(df, filename, columns=...)`.
                    """)
            else:
                drop_column_message = ''

            self._append_issue(
                category='Table too wide',
                issue=dedent_triple_quote_str("""
                    Here is the created table:

                    ```latex
                    {table}
                    ```
                    I tried to compile it, but the table is too wide. 
                    """).format(filename=self.filename, table=latex),
                instructions="Please change the code to make the table narrower. "
                             "Consider any of the following options:\n"
                             + index_note + column_note + drop_column_message + transpose_message,
            )
        else:
            # table is fine
            pass

    CHOICE_OF_CHECKS = CompilationDfContentChecker.CHOICE_OF_CHECKS | {
        check_compilation_and_get_width: True,
    }


""" CONTENT FOR DISPLAY-ITEM STEP """


@dataclass
class SecondContentChecker(BaseContentDfChecker):

    UNWANTED_LABELS = ['intercept']

    def _is_label_unwanted(self, label):
        if isinstance(label, str):
            return any(label.lower() in unwanted_label for unwanted_label in self.UNWANTED_LABELS)

    def check_for_unwanted_labels_in_x_or_y(self):
        for xy in ['x', 'y']:
            labels = self.get_x_labels() if xy == 'x' else self.get_y_labels()
            unwanted_labels = [label for label in labels if self._is_label_unwanted(label)]
            if unwanted_labels:
                nice_unwanted_labels = NiceList(unwanted_labels, wrap_with='"')
                self._append_issue(
                    category='Atypical choice of data to present',
                    issue=f'The {self.table_or_figure} includes data of {nice_unwanted_labels}.',
                    instructions=dedent_triple_quote_str(f"""
                        Including {nice_unwanted_labels} in a scientific {self.table_or_figure} is not common.
                        Please consider removing the {nice_unwanted_labels}.
                        """),
                    forgive_after=2,
                )

    CHOICE_OF_CHECKS = BaseContentDfChecker.CHOICE_OF_CHECKS | {
        check_for_unwanted_labels_in_x_or_y: True,
    }


@dataclass
class TableSecondContentChecker(SecondContentChecker):
    func: Callable = df_to_latex

    def check_for_repetitive_value_in_column(self):
        for icol in range(self.df.shape[1]):
            column_label = self.df.columns[icol]
            data = self.df.iloc[:, icol]
            if is_containing_p_value(data):
                continue
            try:
                data_unique = data.unique()
            except Exception:  # noqa
                data_unique = None
            if data_unique is not None and len(data_unique) == 1 and len(data) > 5:
                data0 = data.iloc[0]
                # check if the value is a number
                if not isinstance(data0, (int, float, np.number)):
                    pass
                elif round(data0) == data0 and data0 < 10:
                    pass
                else:
                    self._append_issue(
                        category='Same value throughout a column',
                        issue=f'The column "{column_label}" has the same unique value for all rows.',
                        instructions=dedent_triple_quote_str(f"""
                            Please revise the code so that it:
                            * Finds the unique values (use `{column_label}_unique = df["{column_label}"].unique()`)
                            * Asserts that there is only one value. (use `assert len({column_label}_unique) == 1`)
                            * Drops the column from the df (use `df.drop(columns=["{column_label}"])`)
                            * Adds the unique value, {column_label}_unique[0], \t
                            in the {self.table_or_figure} note \t
                            (e.g., `{self.func_name}(..., note=f'For all rows, \t
                            the {column_label} is {{{column_label}_unique[0]}}')`)

                            There is no need to add corresponding comments to the code. 
                            """),
                    )

    CHOICE_OF_CHECKS = SecondContentChecker.CHOICE_OF_CHECKS | {
        check_for_repetitive_value_in_column: True,
    }


@dataclass
class SecondFigureContentChecker(SecondContentChecker):
    func: Callable = df_to_figure

    ODDS_RATIO_TERMS_CAPS = [('odds ratio', False), ('OR', True)]

    @property
    def axis_parameters(self) -> AxisParameters:
        return self.intermediate_results.get('axis_parameters')

    def check_log_scale_for_odds_ratios(self):
        """
        Odds ratios should typically be plotted on a log scale.
        Check if the x or y label contains the term "odds ratio".
        """
        for axis in ['x', 'y']:
            label = self.axis_parameters.xlabel if axis == 'x' else self.axis_parameters.ylabel
            is_log = self.axis_parameters.xscale == 'log' if axis == 'x' else self.axis_parameters.yscale == 'log'
            if label is not None and is_log is not True:
                for term, is_caps in self.ODDS_RATIO_TERMS_CAPS:
                    modified_label = label.lower() if not is_caps else label
                    if term in modified_label:
                        self._append_issue(
                            category='Plotting odds ratios',
                            issue=f'The {axis}-axis label contains the term "{term}". Are you plotting odds ratios?\n'
                                  f'If so, odds ratios are typically shown on a log scale; ',
                            instructions=f'Consider using a log scale for the {axis}-axis (setting `log{axis}=True`).',
                            forgive_after=1,
                        )
                        break

    CHOICE_OF_CHECKS = SecondContentChecker.CHOICE_OF_CHECKS | {
        check_log_scale_for_odds_ratios: True,
    }


""" ANNOTATION """


@dataclass
class AnnotationDfChecker(BaseContentDfChecker):
    stop_after_first_issue: bool = False

    UN_ALLOWED_CHARS = [
        # ('_', 'underscore'),
        ('^', 'caret'),
        ('{', 'curly brace'),
        ('}', 'curly brace')
    ]

    @property
    def width(self):
        return self.intermediate_results.get('width')

    @property
    def is_narrow(self):
        return isinstance(self.width, float) and self.width < 0.8

    def check_for_unallowed_characters_in_labels(self):
        any_issues = False
        for char, char_name in self.UN_ALLOWED_CHARS:
            for is_row in [True, False]:
                if is_row:
                    labels = extract_df_row_labels(self.df, with_title=True, string_only=True)
                    index_or_columns = 'index'
                else:
                    labels = extract_df_column_labels(self.df, with_title=True, string_only=True)
                    index_or_columns = 'columns'
                unallowed_labels = sorted([label for label in labels if char in label])
                if unallowed_labels:
                    any_issues = True
                    self._append_issue(
                        category=f'The df row/column labels contain un-allowed characters',
                        issue=dedent_triple_quote_str(f"""
                            The "{self.filename}" has {index_or_columns} labels containing \t
                            the character "{char}" ({char_name}), which is not allowed.
                            Here are the problematic {index_or_columns} labels:
                            {unallowed_labels}
                            """),
                        instructions=dedent_triple_quote_str(f"""
                            Please revise the code to map these {index_or_columns} labels to new names \t
                            that do not contain the "{char}" characters. Spaces are allowed.

                            Doublecheck to make sure your code uses `df.rename({index_or_columns}=...)` \t
                            with the `{index_or_columns}` argument set to a dictionary mapping the old \t
                            {index_or_columns} names to the new ones.
                            """)
                    )
        return any_issues  # we don't want additional issues if working on the underscores

    def check_glossary(self):
        axes_labels = extract_df_axes_labels(self.df, with_title=False, string_only=True)
        abbr_labels = [label for label in axes_labels if is_unknown_abbreviation(label)]
        glossary = self.glossary or {}
        un_mentioned_abbr_labels = sorted([label for label in abbr_labels if label not in glossary])
        glossary_keys_not_in_df = sorted([label for label in glossary if label not in axes_labels])

        if un_mentioned_abbr_labels or glossary_keys_not_in_df:
            glossary_msg = f"The `glossary` argument of `{self.func_name}` for '{self.filename}' "
            if self.glossary:
                glossary_msg += f"includes the following keys:\n{list(self.glossary.keys())}\n\n"
            else:
                glossary_msg += "is not provided.\n\n"
            axes_labels_msg = f'The corresponding df includes the following labels:\n{axes_labels}\n\n'
            if glossary_keys_not_in_df:
                glossary_keys_not_in_df_msg = dedent_triple_quote_str(f"""
                    **Provided glossary keys not matching any of the df labels:**
                    The glossary includes the following keys that are not in the df:
                    {glossary_keys_not_in_df}
                    The glossary keys should be a subset of the df labels.
                    Please remove/replace these stranded keys. \t
                    (or, alternatively, rename the relevant df labels to match these stranded glossary keys).
                    """)
            else:
                glossary_keys_not_in_df_msg = ""

            if un_mentioned_abbr_labels:
                un_mentioned_abbr_labels_msg = dedent_triple_quote_str(f"""\n
                    **Abbreviated df labels that are not mentioned in the glossary:**
                    The df includes the following abbreviated labels that are not defined in the glossary:
                    {un_mentioned_abbr_labels}
                    Please add these missing abbreviated labels to the `glossary`.
                    """)
                if self.is_narrow:
                    un_mentioned_abbr_labels_msg += dedent_triple_quote_str(f"""
                        Alternatively, since the {self.table_or_figure} is not too wide, you can also replace \t
                        these undefined abbreviated labels with their full names in the dataframe itself.
                        """)
            else:
                un_mentioned_abbr_labels_msg = ""

            self._append_issue(
                category='Displayitem glossary',
                issue=f'{glossary_msg}{axes_labels_msg}{glossary_keys_not_in_df_msg}{un_mentioned_abbr_labels_msg}',
                instructions=dedent_triple_quote_str(f"""\n
                    As a reminder: you can also use the `note` argument to add \t
                    information that is related to the f"{self.table_or_figure} as a whole, \t
                    rather than to a specific label.
                    """)
            )

    def _create_displayitem_caption_label_issue(self, issue: str):
        self._append_issue(
            category='Problem with displayitem caption',
            issue=issue,
            instructions=dedent_triple_quote_str("""
                Please revise the code making sure all displayitems are created with a caption.
                Use the arguments `caption` of `df_to_latex` or `df_to_figure`.
                Captions should be suitable for tables/figures of a scientific paper.
                In addition, you can add:
                - an optional note for further explanations (use the argument `note`)
                - a glossary mapping any abbreviated row/column labels to their definitions \t
                (use the argument `glossary` argument). 
                """)
        )

    def _check_caption_or_note(self, text: Optional[str], item_name: str = 'caption', is_required: bool = True):
        forbidden_starts: Tuple[str, ...] = ('Figure', 'Table')
        if text is None:
            if is_required:
                self._create_displayitem_caption_label_issue(
                    f'The {self.table_or_figure} does not have a {item_name}.')
            else:
                return
        else:
            for forbidden_start in forbidden_starts:
                if text.startswith(forbidden_start):
                    self._create_displayitem_caption_label_issue(
                        f'The {item_name} of the {self.table_or_figure} should not start with "{forbidden_start}".')
            if '...' in text:
                self._create_displayitem_caption_label_issue(
                    f'The {item_name} of the {self.table_or_figure} should not contain "..."')
            if re.search(pattern=r'<.*\>', string=text):
                self._create_displayitem_caption_label_issue(
                    f'The {item_name} of the {self.table_or_figure} should not contain "<...>"')

    def check_note(self):
        self._check_caption_or_note(self.note, item_name='note', is_required=False)

    def check_caption(self):
        self._check_caption_or_note(self.caption, item_name='caption', is_required=True)

    def check_note_is_different_than_caption(self):
        note, caption = self.note, self.caption
        if note is not None and caption is not None and (
                note.lower() in caption.lower() or caption.lower() in note.lower()):
            self._create_displayitem_caption_label_issue(
                f'The note of the {self.table_or_figure} should not be the same as the caption.\n'
                'Notes are meant to provide additional information, not to repeat the caption.')

    CHOICE_OF_CHECKS = BaseContentDfChecker.CHOICE_OF_CHECKS | {
        check_for_unallowed_characters_in_labels: True,
        check_glossary: True,
        check_note: True,
        check_caption: True,
        check_note_is_different_than_caption: True,
    }


class FigureAnnotationDfChecker(AnnotationDfChecker):
    func: Callable = df_to_figure

    @property
    def axis_parameters(self) -> AxisParameters:
        return self.intermediate_results.get('axis_parameters')

    def check_if_numeric_axes_have_labels(self):
        """
        Check if all axes with numeric labels have labels.
        Return an error message if not.
        """
        axis_parameters = self.axis_parameters
        x_numeric = axis_parameters.is_x_axis_numeric()
        y_numeric = axis_parameters.is_y_axis_numeric()
        msgs = []
        if x_numeric and not axis_parameters.xlabel:
            msgs.append('The x-axis is numeric, but it does not have a label. Use `xlabel=` to add a label.')
        if y_numeric and not axis_parameters.ylabel:
            msgs.append('The y-axis is numeric, but it does not have a label. Use `ylabel=` to add a label.')
        if msgs:
            msg = 'All axes with numeric labels must have labels.\n' + '\n'.join(msgs)
            self._append_issue(
                issue=msg,
                instructions=dedent_triple_quote_str("""
                    Please revise the code to add labels.
                    Use the arguments `xlabel` and `ylabel` of `df_to_figure`.
                    """),
            )

    CHOICE_OF_CHECKS = AnnotationDfChecker.CHOICE_OF_CHECKS | {
        check_if_numeric_axes_have_labels: True,
    }


""" FILE CONTINUITY """


@dataclass
class ContinuityDfChecker(BaseContentDfChecker):
    DEFAULT_CATEGORY = 'File continuity'

    def check_for_file_continuity(self):
        prior_filename = self.df.get_prior_filename()
        if prior_filename is None:
            self._append_issue(
                issue="The df should be created from a previous df (not from scratch).\n"
                      "Only use inplace operations on the df loaded from the previous step.",
            )
            return
        should_be_filename = prior_filename + '_formatted'
        if self.filename != should_be_filename:
            self._append_issue(
                issue=dedent_triple_quote_str(f"""
                    The file name of the loaded df was "{prior_filename}".
                    The current file name should be "{should_be_filename}" (instead of "{self.filename}").
                    """),
            )

    CHOICE_OF_CHECKS = BaseContentDfChecker.CHOICE_OF_CHECKS | {
        check_for_file_continuity: True,
    }


""" RUN CHECKERS """


def create_and_run_chain_checker_from_list_info_df(checkers: List[Type[BaseContentDfChecker]],
                                                   df: InfoDataFrameWithSaveObjFuncCall, **k
                                                   ) -> Tuple[RunIssues, Dict[str, Any]]:
    func_call = df.get_func_call()
    func, args, kwargs = func_call
    filename = func_call.filename
    return create_and_run_chain_checker(checkers, df=df, func=func, filename=filename, kwargs=kwargs, **k)


def check_df_to_figure_analysis(df: InfoDataFrameWithSaveObjFuncCall, **k) -> RunIssues:
    checkers = [
        FigureSyntaxDfChecker,
        FigureDfContentChecker,
        FigureCompilationDfContentChecker,  # This will actually create the png file
    ]
    return create_and_run_chain_checker_from_list_info_df(checkers, df=df, **k)[0]


def check_df_to_latex_analysis(df: InfoDataFrameWithSaveObjFuncCall, **k) -> RunIssues:
    checkers = [
        TableSyntaxDfChecker,
        TableDfContentChecker,
    ]
    return create_and_run_chain_checker_from_list_info_df(checkers, df=df, **k)[0]


def check_df_to_figure_displayitems(df: InfoDataFrameWithSaveObjFuncCall, **k) -> RunIssues:
    checkers = [
        FigureSyntaxDfChecker,
        FigureDfContentChecker,
        FigureCompilationDfContentChecker,  # Create png. Stores annotation messages in intermediate_results
        ContinuityDfChecker,
        SecondFigureContentChecker,
        FigureAnnotationDfChecker,
    ]
    return create_and_run_chain_checker_from_list_info_df(checkers, df=df, **k)[0]


def check_df_to_latex_displayitems(df: InfoDataFrameWithSaveObjFuncCall, **k) -> RunIssues:
    checkers = [
        TableSyntaxDfChecker,
        TableDfContentChecker,
        ContinuityDfChecker,
        TableSecondContentChecker,
        TableCompilationDfContentChecker,
        AnnotationDfChecker,
    ]
    return create_and_run_chain_checker_from_list_info_df(checkers, df=df, **k)[0]


def check_analysis_df(df: InfoDataFrameWithSaveObjFuncCall, **k) -> RunIssues:
    func, args, kwargs = df.get_func_call()
    if func == df_to_figure:
        return check_df_to_figure_analysis(df, **k)
    elif func == df_to_latex:
        return check_df_to_latex_analysis(df, **k)
    else:
        raise ValueError(f"func should be either df_to_figure or df_to_latex, not {func}")


def check_displayitem_df(df: InfoDataFrameWithSaveObjFuncCall, **k) -> RunIssues:
    func = df.get_func_call().func
    if func == df_to_figure:
        return check_df_to_figure_displayitems(df, **k)
    elif func == df_to_latex:
        return check_df_to_latex_displayitems(df, **k)
    else:
        raise ValueError(f"func should be either df_to_figure or df_to_latex, not {func}")


""" FILENAME CHECKERS """


def get_issue_if_filename_not_legit_df_tag(filename: str, func: Callable = None) -> Optional[RunIssue]:
    """
    check if the filename is in the correct format
    """
    checker = SyntaxDfChecker(filename=filename, func=func)
    checker.check_filename()
    return checker.issues[0] if checker.issues else None


def raise_issue_if_filename_not_legit_df_tag(filename: str, func: Callable = None):
    if not isinstance(filename, str):
        raise ValueError(f"filename should be a string, not {type(filename)}")
    issue = get_issue_if_filename_not_legit_df_tag(filename, func)
    if issue:
        raise issue



"""
This file contains the primary list of actions and the dict of conversations.

The action list allows a formal record of all the actions that were performed on the conversations.
Using the actions, we can replay the entire conversation history.

The dict of conversations is used to keep track of conversations by name, as they are created and modified by actions.
"""

from __future__ import annotations

import pickle
from pathlib import Path
from typing import Union, Dict, List, Optional, Set
from dataclasses import dataclass, field

from data_to_paper.utils.print_to_file import print_and_log
from data_to_paper.base_cast import Agent
from .conversation import Conversation


@dataclass(frozen=True)
class Action:
    """
    Base class for actions performed on a Conversation, Stage, or Cast.
    """
    should_print: bool = True

    def _pretty_attrs(self) -> str:
        return ''

    def pretty_repr(self, is_color: bool = True) -> str:
        s = ''
        s += f'{type(self).__name__}'
        if self._pretty_attrs():
            s += f'({self._pretty_attrs()})'
        return s

    def apply(self):
        """
        Apply the action.
        """
        pass


@dataclass(frozen=True)
class Conversations(Dict[str, Conversation]):
    """
    a dict containing all managed conversations, by name.
    """

    def get_conversation(self, conversation_name: str) -> Optional[Conversation]:
        """
        Return the conversation with the provided name, or None if no such conversation exists.
        """
        return self.get(conversation_name, None)

    def get_or_create_conversation(self, conversation_name: str, participants: Set[Agent] = None) -> Conversation:
        if conversation_name in self:
            return self[conversation_name]
        conversation = Conversation(conversation_name=conversation_name, participants=participants)
        self[conversation.conversation_name] = conversation
        return conversation

    def get_new_conversation_name(self, prefix: str = 'conversation') -> str:
        """
        Return a new conversation name that is not already in use.
        """
        i = 0
        while True:
            conversation_name = f'{prefix} ({i})'
            if conversation_name not in self:
                return conversation_name
            i += 1


@dataclass(frozen=True)
class Actions(List[Action]):
    """
    a list of actions applied to conversations by order in which actions were applied.
    """

    abbreviate_repeated_printed_content: bool = True

    def apply_action(self, action: Action, is_color: bool = True,
                     should_append: bool = True):
        if action.should_print:
            from .conversation_actions import AppendMessage
            if self.abbreviate_repeated_printed_content \
                    and isinstance(action, AppendMessage) \
                    and action.message.content in self.get_all_message_contents(only_printed=True):
                s = action.pretty_repr(is_color=is_color, abbreviate_content=True)
                s_bw = action.pretty_repr(is_color=False, abbreviate_content=True)
            else:
                s = action.pretty_repr(is_color=is_color)
                s_bw = action.pretty_repr(is_color=False)
            if s:
                print_and_log(s_bw, text_in_color=s, end='\n\n')
        if should_append:
            self.append(action)
        action.apply()

    def save_actions_to_file(self, file_path: Union[str, Path]):
        """
        Save the primary list of actions to a json file.
        """
        with open(file_path, 'wb') as f:
            pickle.dump(self, f)

    def load_actions_from_file(self, file_path: Union[str, Path]):
        """
        Load a list of actions from a json file.
        """
        with open(file_path, 'rb') as f:
            self.clear()
            self.extend(pickle.load(f))
        return self

    def get_actions_for_conversation(self, conversation_name: str) -> List[Action]:
        """
        Return a list of actions that were applied to the conversation with the provided name.
        """
        from .conversation_actions import ChangeMessagesConversationAction
        return [action for action in self if
                isinstance(action, ChangeMessagesConversationAction) and action.conversation_name == conversation_name]

    def get_all_message_contents(self, only_printed: bool = False) -> List[str]:
        """
        Return a list of all message contents.
        """
        from .conversation_actions import AppendMessage
        return [action.message.content for action in self
                if isinstance(action, AppendMessage) and (not only_printed or
                                                          (action.should_print and
                                                           action.should_add_to_conversation()))
                ]


@dataclass(frozen=True)
class ActionsAndConversations:
    actions: Actions = field(default_factory=Actions)
    conversations: Conversations = field(default_factory=Conversations)




def is_lower_eq(value, threshold):
    """
    Check if `value` is less than or equal to `threshold`.
    `threshold` of `None` means no threshold.
    """
    return threshold == None or value <= threshold  # noqa. == to work with Mutable



from PySide6.QtCore import QEvent, Qt
from PySide6.QtWidgets import QApplication, QPlainTextEdit
from unidecode import unidecode


class PlainTextPasteTextEdit(QPlainTextEdit):
    def __init__(self, parent=None):
        super().__init__(parent)

    def keyPressEvent(self, event):
        # Check for Ctrl+V (Paste shortcut)
        if event.key() == Qt.Key_V and (event.modifiers() & Qt.ControlModifier):
            self.handle_paste()
        else:
            # Handle regular key press
            text = event.text()
            if text:
                utf8_text = unidecode(text)
                # Insert the converted text manually
                cursor = self.textCursor()
                cursor.insertText(utf8_text)
            else:
                # Call parent method if no text to process (e.g., for non-character keys)
                super().keyPressEvent(event)

    def handle_paste(self):
        # Get the text from the clipboard
        clipboard = QApplication.clipboard()
        text = clipboard.text()

        # Convert the clipboard text to UTF-8 using unidecode
        utf8_text = unidecode(text)

        # Insert the converted text into the editor
        self.insertPlainText(utf8_text)

    def event(self, event):
        # Intercept the clipboard paste event
        if event.type() == QEvent.Clipboard:
            self.handle_paste()
            return True
        return super().event(event)

    def setPlainText(self, text):
        # Convert the text to UTF-8 using unidecode
        return super().setPlainText(unidecode(text))

    def toPlainText(self):
        # Get the text and convert it to UTF-8 using unidecode
        return unidecode(super().toPlainText())



from data_to_paper.conversation.stage import Stage


class DemoStages(Stage):
    DATA = ("Get Data", False)
    GOAL = ("Set Goal", False)
    CODE = ("Write Code", True)
    WRITING = ("Write Abstract", True)
    COMPILE = ("Compile pdf", False)



import re
from typing import Optional, Dict, Tuple, Callable
from functools import partial

import colorama
from pygments.formatters.html import HtmlFormatter
from pygments.lexers import PythonLexer, JsonLexer
from pygments.lexer import RegexLexer
from pygments.formatters import Terminal256Formatter
from pygments.styles import get_style_by_name
from pygments import highlight, token
from typing import List

from data_to_paper.latex.latex_to_html import convert_latex_to_html
from data_to_paper.env import CHOSEN_APP

from .formatted_sections import FormattedSections
from .text_formatting import wrap_string, wrap_as_block

COLORS_TO_LIGHT_COLORS = {
    colorama.Fore.BLACK: colorama.Fore.LIGHTBLACK_EX,
    colorama.Fore.RED: colorama.Fore.LIGHTRED_EX,
    colorama.Fore.GREEN: colorama.Fore.LIGHTGREEN_EX,
    colorama.Fore.YELLOW: colorama.Fore.LIGHTYELLOW_EX,
    colorama.Fore.BLUE: colorama.Fore.LIGHTBLUE_EX,
    colorama.Fore.MAGENTA: colorama.Fore.LIGHTMAGENTA_EX,
    colorama.Fore.CYAN: colorama.Fore.LIGHTCYAN_EX,
    colorama.Fore.WHITE: colorama.Fore.LIGHTWHITE_EX,
    "": "",
}

style = get_style_by_name("monokai")
terminal_formatter = Terminal256Formatter(style=style)

if CHOSEN_APP == 'pyside':
    html_code_formatter = HtmlFormatter(style=style)
else:
    html_code_formatter = HtmlFormatter(style=style, cssclass="code_highlight", prestyles="margin-left: 1.5em;")


class CSVLexer(RegexLexer):
    name = 'CSV'
    aliases = ['csv']
    filenames = ['*.csv']

    tokens = {
        'root': [
            # Matches integers, floats, and numbers in scientific notation with optional leading +/-
            (r'[-+]?\b[0-9]+(\.[0-9]*)?([eE][-+]?[0-9]+)?\b', token.Number.Float),
            (r'0[oO]?[0-7]+', token.Number.Oct),  # Octal literals
            (r'0[xX][a-fA-F0-9]+', token.Number.Hex),  # Hexadecimal literals
            (r'\b[0-9]+[jJ]\b', token.Number),  # Complex numbers
            (r'.', token.Token.Name),
        ]
    }


def python_to_highlighted_html(code_str: str) -> str:
    return highlight(code_str, PythonLexer(), html_code_formatter)


def output_to_highlighted_html(output_str: str) -> str:
    return highlight(output_str, CSVLexer(), html_code_formatter)


def python_to_highlighted_text(code_str: str, color: str = '', label: Optional[str] = None) -> str:
    if color:
        return highlight(code_str, PythonLexer(), terminal_formatter)
    else:
        return code_str


def json_to_highlighted_html(json_str: str) -> str:
    return highlight(json_str, JsonLexer(), html_code_formatter)


def json_to_highlighted_text(json_str: str, color: str = '', label: Optional[str] = None) -> str:
    if color:
        return highlight(json_str, JsonLexer(), terminal_formatter)
    else:
        return json_str


def demote_html_headers(html, demote_by=1):
    """
    Demote HTML header tags by a specified number of levels.
    """
    if demote_by < 1:
        return html  # Return the original html if demote_by is less than 1

    # Function to replace header tags
    def replace_tag(match):
        current_level = int(match.group(1))
        new_level = min(current_level + demote_by, 6)  # HTML only supports headers up to <h6>
        return f'<h{new_level}>{match.group(2)}</h{new_level}>'

    # Regex to find headers and replace
    return re.sub(pattern=r'<h([1-6])>(.*?)</h\1>', repl=replace_tag, string=html)


def md_to_html(md):
    """
    Convert markdown to HTML while managing newline characters appropriately after headers.
    """
    html_lines = []
    for line in md.split('\n'):
        if re.match(pattern=r'^#{1,5} ', string=line):
            header_level = len(line.split(' ')[0])
            html_line = f'<h{header_level}>{line[header_level + 1:]}</h{header_level}>'
            if html_lines and html_lines[-1] == '<br>':
                html_lines.pop()
        # TODO: The bullet formatting looks ugly. Disable it for now.
        # elif line.startswith('- '):
        #     html_line = f'<li>- {line[2:]}</li>'
        # elif line.startswith('* '):
        #     html_line = f'<li>* {line[2:]}</li>'
        elif not line.strip():
            html_line = '<br>'
        else:
            html_line = line + '<br>'
        html_lines.append(html_line)
    md = '\n'.join(html_lines)

    # Convert bold:
    md = re.sub(pattern=r'\*\*(.*?)\*\*', repl=r'<b>\1</b>', string=md)

    # Convert italics:
    md = re.sub(pattern=r'\*(.*?)\*', repl=r'<i>\1</i>', string=md)

    # Convert code, flanked by backticks:
    md = re.sub(pattern=r'`([^`]+)`', repl=r'<span class="codeline">\1</span>', string=md)
    return md


def text_to_html(text: str, label: Optional[str] = None,
                 from_md: bool = False, css_class: Optional[str] = 'markdown') -> str:
    # strip newlines from the right end of the text:
    text = text.strip('\n')
    text = _escape_html(text)
    if from_md:
        html = md_to_html(text)
    else:
        html = text.replace('\n', '<br>')
    if css_class is not None:
        html = f'<div class="{css_class}">{html}</div>'
    return html


def colored_text(text: str, color: str, is_color: bool = True, is_light: bool = False) -> str:
    if not is_color or color == '':
        return text
    if is_light:
        color = COLORS_TO_LIGHT_COLORS[color]
    return color + text + colorama.Style.RESET_ALL


def red_text(text: str, is_color: bool = True) -> str:
    return colored_text(text, colorama.Fore.RED, is_color)


def _colored_block(text: str, label: Optional[str], color: str, with_tags: bool = True,
                   is_color: bool = True, is_light: bool = False) -> str:
    if with_tags and label:
        text = wrap_as_block(text, label)
    return colored_text(text, color, is_color, is_light)


_light_colored_block = partial(_colored_block, is_light=True)
_light_colored_block_no_tags = partial(_light_colored_block, with_tags=False)


def _escape_html(text: str) -> str:
    return text.replace('<', '&lt;').replace('>', '&gt;')


def get_pre_html_format(text,
                        color: str = None,
                        font_style: str = None,  # normal, italic, oblique
                        font_size: int = None,
                        font_weight: str = None,  # normal, bold
                        font_family: str = None):
    text = _escape_html(text)
    s = '<pre style="white-space: pre-wrap;'
    if color:
        s += f'color: {color};'
    if font_style:
        s += f'font-style: {font_style};'
    if font_size:
        s += f'font-size: {font_size}px;'
    if font_weight:
        s += f'font-weight: {font_weight};'
    if font_family:
        s += f'font-family: {font_family};'
    s += '">'
    return s + text + '</pre>'


def _block_to_html(text: str, label: Optional[str], with_tags: bool = True, **kwargs) -> str:
    if with_tags and label is not None:
        text = wrap_as_block(text, label)
    return text_to_html(text, css_class='tripled_quote')


def identity(text) -> str:
    return text


NORMAL_FORMATTERS = (_colored_block, text_to_html)
BLOCK_FORMATTERS = (_light_colored_block, _block_to_html)

TAGS_TO_FORMATTERS: Dict[Optional[str], Tuple[Callable, Callable]] = {
    None: NORMAL_FORMATTERS,
    '': BLOCK_FORMATTERS,
    'markdown': (_light_colored_block, partial(text_to_html, from_md=True, label=None)),
    'md': (_light_colored_block, partial(text_to_html, from_md=True, label=None)),
    'python': (python_to_highlighted_text, python_to_highlighted_html),
    'output': (_light_colored_block, output_to_highlighted_html),
    'html': (_light_colored_block, identity),
    'header': (_light_colored_block_no_tags, partial(get_pre_html_format, color='#FF0000', font_size=12)),
    'latex': (_light_colored_block, convert_latex_to_html),
    'error': (partial(_light_colored_block_no_tags, color=colorama.Fore.RED),
              partial(text_to_html, css_class="runtime_error")),
    'json': (json_to_highlighted_text, json_to_highlighted_html),
    'system': (_light_colored_block_no_tags, NotImplemented),
    'comment': (_light_colored_block_no_tags, NotImplemented),
}


def is_text_md(text: str) -> bool:
    return text.strip().startswith('#')


def format_text_with_code_blocks(text: str, text_color: str = '', from_md: Optional[bool] = None,
                                 width: Optional[int] = 150, is_html: bool = False,
                                 do_not_format: List[str] = None) -> str:
    if from_md is None:
        from_md = is_text_md(text)
    do_not_format = do_not_format or []
    s = ''
    formatted_sections = FormattedSections.from_text(text)
    for formatted_section in formatted_sections:
        label, section, is_complete = formatted_section.to_tuple()
        if label is not None:
            label = label.lower()
        if not is_complete:
            formatters = NORMAL_FORMATTERS
            section = formatted_section.to_text()
        else:
            if label in do_not_format:
                formatters = BLOCK_FORMATTERS
            else:
                formatters = TAGS_TO_FORMATTERS.get(label, BLOCK_FORMATTERS)
        formatter = formatters[is_html]
        if is_html:
            if formatter == text_to_html:
                s += formatter(section, from_md=from_md)
            elif formatter == _block_to_html:
                s += formatter(section, label=label)
            else:
                s += formatter(section)
        else:
            section = wrap_string(section, width=width)
            s += formatter(section, color=text_color, label=label)
    return s



import os
import traceback

from pathlib import Path

from data_to_paper.utils.mutable import Flag

MODULE_NAME = 'script_to_run'
module_filename = MODULE_NAME + ".py"


IS_CHECKING = Flag(False)


def is_filename_gpt_code(filename: str) -> bool:
    """
    Check if the filename is a gpt code filename or a test filename.
    """
    folder = Path(filename).parent
    filename = os.path.basename(filename)
    # TODO: this is a hack. Need to define the file of the gpt script dynamically.
    return filename == module_filename \
        or str(folder).replace('\\', '/').endswith('data_to_paper/scripts')


def is_filename_test(filename: str) -> bool:
    filename = os.path.basename(filename)
    return filename.startswith('test_')


def get_gpt_module_frames(tb: traceback.StackSummary) -> list:
    frames = [t for t in tb if is_filename_gpt_code(t.filename)]
    if len(frames):
        return frames
    frames = [t for t in tb if is_filename_test(t.filename)]
    return frames


def is_called_from_user_script(offset: int = 3) -> bool:
    """
    Check if the code is called from user script.
    """
    if IS_CHECKING:
        return False
    with IS_CHECKING.temporary_set(True):
        tb = traceback.extract_stack()  # this can lead to import and invoke recursion (thereby IS_CHECKING))
        filename = tb[-offset].filename
        return is_filename_gpt_code(filename) or is_filename_test(filename)



from dataclasses import dataclass, field
from functools import partial
from typing import Dict, List, Collection, Optional, Any

from data_to_paper.env import PAUSE_AFTER_LITERATURE_SEARCH, JSON_MODE, SCHOLAR_SERVER

from data_to_paper.text import dedent_triple_quote_str, word_count
from data_to_paper.utils.nice_list import NiceDict, NiceList
from data_to_paper.utils.print_to_file import print_and_log_red
from data_to_paper.servers.semantic_scholar import (
    SEMANTIC_SCHOLAR_EMBEDDING_SERVER_CALLER,
)

from data_to_paper.interactive import PanelNames

from .request_python_value import (
    PythonDictWithDefinedKeysReviewBackgroundProductsConverser,
)
from .literature_search import (
    LiteratureSearch,
    QueryCitationCollectionProduct,
    LiteratureSearchQueriesProduct,
)


@dataclass
class BaseLiteratureSearchReviewGPT(
    PythonDictWithDefinedKeysReviewBackgroundProductsConverser
):
    json_mode: bool = JSON_MODE
    number_of_papers_per_query: int = 100
    max_reviewing_rounds: int = 0
    literature_search: LiteratureSearch = field(default_factory=LiteratureSearch)
    domains_to_definitions_and_examples = {
        "dataset": {
            "definition": "papers that use the same or similar datasets as in our study",
            "examples": ["The UK-NCD dataset", "covid-19 vaccine efficacy dataset"],
        },
        "questions": {
            "definition": "papers that ask questions similar to our study",
            "examples": [
                "covid-19 vaccine efficacy over time",
                "covid-19 vaccine waning",
            ],
        },
        "background": {
            "definition": "papers that provide background on the overall subject of our study",
            "examples": [
                "SARS-CoV2 spread",
                "covid-19 global impact",
                "covid-19 vaccine",
            ],
        },
        "methods": {
            "definition": "papers that use the same or similar methods as in our study",
            "examples": [
                "covid-19 vaccine efficacy analysis",
                "kaplan-meier survival analysis",
            ],
        },
        "results": {
            "definition": "papers that report results similar to our study",
            "examples": [
                "covid-19 vaccine efficacy",
                "covid-19 vaccine efficacy over time",
                "covid-19 vaccine waning",
            ],
        },
    }
    requested_keys: Collection[str] = (
        "dataset",
        "questions",
    )

    excluded_citation_titles: List[str] = None
    # bibtex ids to remove from the results. Important for removing reproduced paper from citing itself.

    value_type: type = Dict[str, List[str]]
    goal_noun: str = "literature search queries"
    goal_verb: str = "write"

    mission_prompt: str = dedent_triple_quote_str(
        """
        Please write literature-search queries that we can use to search for papers related to our study.

        You would need to compose search queries to identify prior papers covering these {num_scopes} areas:
        {pretty_scopes_to_definitions}

        Return your answer as {your_response_should_be_formatted_as}, \t
        where the keys are the {num_scopes} areas noted above, \t
        and the values are lists of query string.

        Each individual query should be a string with up to 5-10 words. 

        For example, for a study reporting waning of the efficacy of the covid-19 BNT162b2 vaccine based on analysis \t
        of the "United Kingdom National Core Data (UK-NCD)", the queries could be:
        ```{python_or_json}
        {pretty_scopes_to_examples}
        ```

        Your response should be formatted as {your_response_should_be_formatted_as}  
        """
    )

    @property
    def chosen_domains_to_definitions_and_examples(self) -> Dict[str, Dict[str, str]]:
        return {
            key: self.domains_to_definitions_and_examples[key]
            for key in self.requested_keys
        }

    @property
    def pretty_scopes_to_definitions(self) -> str:
        return "\n".join(
            [
                f'"{scope}": {definition_and_examples["definition"]}'
                for scope, definition_and_examples in self.chosen_domains_to_definitions_and_examples.items()
            ]
        )

    @property
    def pretty_scopes_to_examples(self) -> str:
        nice = partial(NiceList, wrap_with='"', prefix="[", suffix="]", separator=", ")
        return (
            "{\n"
            + "\n".join(
                [
                    f'    "{scope}": {nice(definition_and_examples["examples"])}'
                    for scope, definition_and_examples in self.chosen_domains_to_definitions_and_examples.items()
                ]
            )
            + "\n}"
        )

    @property
    def num_scopes(self) -> int:
        return len(self.requested_keys)

    def get_title(self) -> Optional[str]:
        """
        Returns the title of the paper we are writing.
        None to skip building embedding vector.
        """
        return None

    def get_abstract(self) -> Optional[str]:
        """
        Returns the abstract of the paper we are writing.
        None to skip building embedding vector.
        """
        return None

    def _check_response_value(self, response_value: dict) -> NiceDict:
        super()._check_response_value(response_value)
        # The queries are 'valid' even if they have too many words:
        self._update_valid_result(response_value)
        too_long_queries = []
        for queries in response_value.values():
            for query in queries:
                if word_count(query) > 10:
                    too_long_queries.append(query)
        if too_long_queries:
            self._raise_self_response_error(
                title="# Too long queries",
                error_message=dedent_triple_quote_str(
                    """
                Queries should be 5-10 word long.

                The following queries are too long:
                {}

                Please return your complete response again, with these queries shortened.
                """
                ).format(
                    NiceList(
                        too_long_queries,
                        wrap_with='"',
                        prefix="",
                        suffix="",
                        separator="\n",
                    )
                ),
            )
        return NiceDict(
            {
                k: NiceList(
                    v,
                    wrap_with='"',
                    prefix="[\n" + " " * 8,
                    suffix="\n" + " " * 4 + "]",
                    separator=",\n" + " " * 8,
                )
                for k, v in response_value.items()
            }
        )

    def _send_html_and_scroll_to_bottom(self, html: str):
        self._app_send_prompt(
            PanelNames.FEEDBACK, html, provided_as_html=True, scroll_to_bottom=True
        )

    def get_literature_search(self) -> LiteratureSearch:
        scopes_to_list_of_queries = self.run_and_get_valid_result()
        literature_search = self.literature_search
        with self._app_temporarily_set_panel_status(
            PanelNames.FEEDBACK, "Querying citations..."
        ):
            html = "<h1>Literature Search</h1>\n"
            if self.get_title() is not None and self.get_abstract() is not None:
                # Requesting embedding vector of our paper
                html += f"<h2>Getting Embedding Vector</h2>\n"
                html += (
                    f'<p>Requesting "{SEMANTIC_SCHOLAR_EMBEDDING_SERVER_CALLER.name}" '
                    f"for the embedding vector of our draft title and abstract....</p>\n"
                )
                self._send_html_and_scroll_to_bottom(html)
                literature_search.embedding_target = (
                    SEMANTIC_SCHOLAR_EMBEDDING_SERVER_CALLER.get_server_response(
                        {
                            "paper_id": "",
                            "title": self.get_title(),
                            "abstract": self.get_abstract(),
                        }
                    )
                )
                html += (
                    "<p>Embedding vector successfully retrieved.</p>\n"
                    "<p>Now we can sort the papers by similarity to our study.</p>\n"
                )
                self._send_html_and_scroll_to_bottom(html)

            html += f"<h2>Querying Citations</h2>\n"
            html += (
                f'<p>Searching "{SCHOLAR_SERVER.server}" '
                f"for papers related to our study in the following areas:</p>\n"
            )
            for scope, queries in scopes_to_list_of_queries.items():
                queries_to_citations = {}
                html += f"<h3>{scope.title()}-related queries:</h3>\n"
                for query in queries:
                    citations = (
                        SCHOLAR_SERVER.get_server_instance().get_server_response(
                            query, rows=self.number_of_papers_per_query
                        )
                    )
                    num_citations = len(citations)
                    html += f'<p><b style="color: #1E90FF;">Query:</b> "{query}".\n'
                    html += f'<br><b style="color: #1E90FF;">Found:</b> {num_citations} citations.</p>\n'
                    self._send_html_and_scroll_to_bottom(html)
                    self.comment(
                        f"\nQuerying {SCHOLAR_SERVER.server}. "
                        f"Found {num_citations} / {self.number_of_papers_per_query} citations. "
                        f'Query: "{query}".'
                    )
                    if self.excluded_citation_titles is not None:
                        excluded_citations = [
                            citation
                            for citation in citations
                            if citation["title"] in self.excluded_citation_titles
                        ]
                        if excluded_citations:
                            print_and_log_red(
                                f"The following citations specified in the excluded citation list were excluded:\n"
                            )
                            for citation in excluded_citations:
                                print_and_log_red(f"{citation}\n\n")
                            citations = QueryCitationCollectionProduct(
                                query=query,
                                value=[
                                    citation
                                    for citation in citations
                                    if citation not in excluded_citations
                                ],
                            )

                    queries_to_citations[query] = citations

                literature_search[scope] = queries_to_citations

        self._app_request_panel_continue(
            PanelNames.FEEDBACK, sleep_for=PAUSE_AFTER_LITERATURE_SEARCH
        )
        return literature_search

    def _update_valid_result(self, valid_result: Any):
        super()._update_valid_result(LiteratureSearchQueriesProduct(value=valid_result))



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Optional, Tuple, Union, Iterable, Type

from data_to_paper.base_products.product import Product, ValueProduct
from data_to_paper.base_steps.converser import Converser
from data_to_paper.base_steps.exceptions import FailedCreatingProductException
from data_to_paper.conversation.message_designation import RangeMessageDesignation, SingleMessageDesignation
from data_to_paper.conversation.stage import Stage
from data_to_paper.env import PAUSE_AT_RULE_BASED_FEEDBACK
from data_to_paper.exceptions import data_to_paperException
from data_to_paper.interactive import PanelNames
from data_to_paper.text.highlighted_text import format_text_with_code_blocks
from data_to_paper.utils.mutable import Flag
from data_to_paper.utils.print_to_file import print_and_log_red
from data_to_paper.utils.replacer import StrOrReplacer, format_value


class Rewind(Enum):
    """
    An enum for the different ways to rewind the conversation upon invalid response.

    For example:

    (1) user initiation prompt

    (2) here is the code
    (3) you have an error

    (4) Here is the revised code
    (5) you have an error in the revised code

    (6) Here is the revised-revised code
    (7) you have an error in the revised-revised code

    REGENERATE: delete (6)-(7)
    RESTART: delete (2)-(7)
    AS_FRESH: delete (2)-(5) and change (6) to look fresh
    ACCUMULATE: do nothing
    AS_FRESH_CORRECTION: delete (4)-(5)
    DELETE_ALL: delete (1)-(7)
    """

    REGENERATE = 'regenerate'
    # regenerate the last response

    RESTART = 'restart'
    # deleted all iterations and get a fresh response

    AS_FRESH = 'as_fresh'
    # delete all previous responses and modify and post the current response as a fresh response

    ACCUMULATE = 'accumulate'
    # just normally add the current response and the feedback

    AS_FRESH_CORRECTION = 'as_fresh_correction'
    # delete any previous erroneous responses except the first one and modify and post the current response as a fresh
    # response

    DELETE_ALL = 'delete_all'
    # delete all previous responses including the original user initiation prompt


class BumpModel(Enum):
    """
    An enum for the different ways to bump the model upon invalid response.
    """

    DO_NOT_BUMP = 'do_not_bump'

    HIGHER_STRENGTH = 'higher_strength'
    # bump the model to a model with higher strength

    HIGHER_CONTEXT = 'higher_context'
    # bump the model to a model with higher context

    @classmethod
    def from_is_higher_context(cls, is_higher_context: bool):
        return cls.HIGHER_CONTEXT if is_higher_context else cls.HIGHER_STRENGTH

    def __bool__(self):
        return self != self.DO_NOT_BUMP


@dataclass
class SelfResponseError(data_to_paperException):
    """
    Exception raised when the response to a request for a latex section is not acceptable.
    """
    error_message: str = None
    title: str = ''
    formatting_instructions: str = None
    rewind: Rewind = None
    bump_model: Optional[BumpModel] = None
    add_iterations: int = 0
    # if positive, we will *reduce* the iteration count so that we have more iterations to correct the response

    def __str__(self):
        return f'SelfResponseError: {self.error_message}'


class NoResponse:
    pass


ExtractedText = Union[str, Iterable[str]]


@dataclass
class ResultConverser(Converser):
    """
    A converser that is designed to extract a result from a conversation.
    response        ---> extracted_text ---> valid_result --> Product
                                                                /
    fresh_response  <--- extracted_text <--- valid_result  <---
    """

    performer: str = 'scientist'

    # goal_noun: the desired output of the conversation (expressed as a singular noun).
    goal_noun: str = 'one-paragraph summary on the solar system'

    # goal_verb: a verb applied to achieve the goal, like 'write', 'draw', 'build', 'code', etc.
    goal_verb: str = 'write'

    # *** Properties that are more generic (adjust only if needed) ***

    system_prompt: str = "You are a {performer} who needs to {goal_verb} {goal_noun}."

    mission_prompt: str = "Please {goal_verb} {goal_noun}."

    max_valid_response_iterations: int = 6

    your_response_should_be_formatted_as: str = ""

    formatting_instructions_for_feedback: str = \
        "Remember, your response should be formatted as {your_response_should_be_formatted_as}"

    default_rewind_for_result_error: Rewind = Rewind.AS_FRESH
    # Can be any of the Rewind options. In particular:
    # ACCUMULATE: just add and accumulate the response and the error message.
    # AS_FRESH: reformat the response as fresh and repost as if it was the first response.
    # AS_FRESH_CORRECTION: reformat the response as fresh and repost as if it was the second response.
    #                      (the first response is left as is; this is good for responses that include chain-of-thought)

    rewind_after_getting_a_valid_response: Optional[Rewind] = None
    # Can be:
    # DELETE_ALL: leave the conversation as if the exchange never happened
    # AS_FRESH: rewind back to right after the user initiation prompt and post the last response as fresh
    # ACCUMULATE (or `None`): do not do anything. the exchange is left as is.

    _conversation_len_before_first_response: int = None
    _self_response_iteration_count: int = 0
    _is_extracting: Flag = field(default_factory=Flag)

    # Output:
    stage: Stage = None
    product_type: Type[ValueProduct] = None  # the type of the product to be generated. If None, non-product result.
    valid_result: Union[Product, Any] = field(default_factory=NoResponse)
    _valid_result_update_count: int = 0

    def get_valid_result_as_html(self) -> str:
        valid_result = self._get_valid_result()
        if isinstance(valid_result, Product):
            return valid_result.as_html(2)
        return format_text_with_code_blocks(self.get_valid_result_as_markdown(), width=None, is_html=True, from_md=True)

    def get_valid_result_as_markdown(self) -> str:
        valid_result = self._get_valid_result()
        if isinstance(valid_result, Product):
            return valid_result.as_formatted_text(2)
        return str(valid_result)

    def _upon_conversation_initiation(self):
        super()._upon_conversation_initiation()
        if self.goal_noun:
            self._app_set_header(self.conversation_name)

    def initialize_conversation_if_needed(self):
        super().initialize_conversation_if_needed()
        self._pre_populate_background()
        self._conversation_len_before_first_response = len(self.conversation)

    def _pre_populate_background(self):
        """
        Add background messages to the two conversations to set them ready for the cycle.
        """
        if self.mission_prompt:
            self.apply_append_user_message(self.mission_prompt, app_panel=PanelNames.MISSION_PROMPT,
                                           editing_title='Revise as needed',
                                           in_field_instructions='Write the mission prompt.')

    @property
    def _has_valid_result(self) -> bool:
        """
        Return whether we have a result.
        """
        return not isinstance(self.valid_result, NoResponse)

    def _update_valid_result(self, valid_result: Union[Product, Any]):
        """
        Update the valid result.
        Should be called when we have a result that is "usable".
        Typically, the method is called after passing all rule-based checks.
        But can also be called when we have a result that passed key rules but not others so is "good enough"
        to be declared as "valid". It will be used in case the improvement iteration count is reached.
        """
        valid_result = self._convert_valid_result_to_product(valid_result)
        if isinstance(valid_result, Product):
            if valid_result.stage is None:
                valid_result.stage = self.stage
        self.valid_result = valid_result
        self._valid_result_update_count += 1
        if self.app:
            if self._has_valid_result:
                html = self.get_valid_result_as_html()
            else:
                html = "No result yet."
            self._app_send_prompt(PanelNames.PRODUCT, html, provided_as_html=True)

    def _raise_self_response_error(self,
                                   title: str,
                                   error_message: StrOrReplacer,
                                   formatting_instructions: StrOrReplacer = None,
                                   missing_end: bool = False,
                                   rewind: Optional[Rewind] = None,
                                   add_iterations: Optional[int] = None,
                                   bump_model: Optional[BumpModel] = None):
        """
        Raise a SelfResponseError with the given error message and instructions for how to rewind the conversation.

        If we are extracting the result:
        Since we could not extract the result, we cannot repost it as fresh.
        Therefore, the default behavior is:
        - if this is the first response, we add the error message (accumulate).
        - otherwise, we regenerate the response. Because:
          if the first response was extractable, we already have example of a correctly formatted response.
          If the first response was not extractable, we already answered with formatting instructions.
        """
        is_first_response = self._conversation_len_before_first_response == len(self.conversation) - 1
        if self._is_extracting:
            if rewind is None:
                if is_first_response:
                    rewind = Rewind.ACCUMULATE
                else:
                    rewind = Rewind.REGENERATE
            if add_iterations is None:
                add_iterations = 0
            if bump_model is None:
                if missing_end:
                    bump_model = BumpModel.HIGHER_CONTEXT
                else:
                    bump_model = BumpModel.DO_NOT_BUMP
        else:
            if rewind is None:
                rewind = self.default_rewind_for_result_error
            if add_iterations is None:
                add_iterations = 0
            if bump_model is None:
                bump_model = BumpModel.DO_NOT_BUMP
        raise SelfResponseError(format_value(self, error_message),
                                title=title, formatting_instructions=formatting_instructions,
                                rewind=rewind, bump_model=bump_model,
                                add_iterations=add_iterations)

    def _convert_response_error_to_error_message(self, response_error: SelfResponseError) -> str:
        """
        Convert the response error to an error message.
        """
        formatting_instructions = response_error.formatting_instructions or self.formatting_instructions_for_feedback
        return format_value(self, response_error.title) + '\n' + format_value(self, response_error.error_message) \
            + '\n' + format_value(self, formatting_instructions)

    """
    Response --> extracted_text --> valid_result --> Product
    """

    def _check_response_and_get_extracted_text(self, response: str) -> ExtractedText:
        """
        # Response --> extracted_text #
        Check the response from self and extract the part(s) that should be used to get the valid result and
        to compose a fresh looking response.
        If there are errors that require self to revise the response, call _raise_self_response_error.
        """
        return response

    def _check_extracted_text_and_update_valid_result(self, extracted_text: ExtractedText):
        """
        # extracted_text --> valid_result #
        Check the extracted_text and extract the needed information into valid_result.
        If we get a result that is "usable", we update it using valid_result, by calling _update_valid_result.
        If there are errors that require self to revise the response, call _raise_self_response_error.
        Normally, we should either _update_valid_result or _raise_self_response_error, but we can do both if we have a
        result that is "good enough" to be declared as "valid", but we are still trying to improve it with further
        rule-based feedback.
        """
        self._update_valid_result(extracted_text)

    def _convert_valid_result_to_product(self, valid_result: Any) -> Union[Product, Any]:
        """
        # valid_result --> Product #
        Convert the valid result to a product.
        """
        if self.product_type is None:
            return valid_result
        return self.product_type(value=valid_result)

    """
    fresh_response <-- extracted_text <-- valid_result <-- Product
    """

    def _convert_extracted_text_to_fresh_looking_response(self, extracted_text: ExtractedText) -> str:
        """
        # fresh_response <-- extracted_text #
        Convert the extracted text to a response that can be posted to the conversation.
        """
        if isinstance(extracted_text, str):
            return extracted_text
        return '\n\n'.join(extracted_text)

    def _convert_valid_result_back_to_extracted_text(self, valid_result: Any) -> ExtractedText:
        """
        # extracted_text <-- valid_result #
        Convert the valid result to an extracted result.
        """
        return str(valid_result)

    def _convert_product_back_to_valid_result(self, product: Union[Product, Any]) -> Any:
        """
        # valid_result <-- Product #
        Convert the product to a valid result.
        """
        if isinstance(product, ValueProduct):
            return product.value
        return product

    def _convert_valid_results_to_fresh_looking_response(self, valid_results: Any) -> str:
        """
        # fresh_response <-- extracted_text <-- valid_result <-- Product #
        Convert the valid results to a response that can be posted to the conversation.
        """
        valid_results = self._convert_product_back_to_valid_result(valid_results)
        extracted_text = self._convert_valid_result_back_to_extracted_text(valid_results)
        fresh_response = self._convert_extracted_text_to_fresh_looking_response(extracted_text)
        return fresh_response

    def _rewind_conversation_to_first_response(self, offset: int = 0, last: int = -1, start: int = None):
        """
        Rewind the conversation to the first response + offset.
        offset=0 means that we delete all messages including the first response.
        """
        if start is None:
            start = self._conversation_len_before_first_response
        self.apply_delete_messages(
            RangeMessageDesignation.from_(start + offset, last))

    def _iterate_until_valid_response(self) -> Tuple[bool, bool]:
        """
        Iterate until we get a valid response from self (return is_converged=True),
        or until we exceed max_valid_response_iterations (return is_converged=False).
        Note that even if we did not get a fully valid response,
        we may still have a "usable" result in self.valid_result (is_new_valid_result=True).
        return (converged, is_new_valid_result)
        """
        self._conversation_len_before_first_response = len(self.conversation)
        self_message = None
        self_response = None
        self._self_response_iteration_count = 0
        initial_valid_result_update_count = self._valid_result_update_count
        is_new_valid_result = False
        while self._self_response_iteration_count < self.max_valid_response_iterations:
            self._self_response_iteration_count += 1
            # to allow starting either before or after the first self response:
            is_preexisting_self_response = True
            try:
                self_response = self.conversation.get_last_response()
                # We are starting after the first self response
            except ValueError:
                # We are starting before the first self response
                is_preexisting_self_response = False

            if self._self_response_iteration_count == 1 and is_preexisting_self_response:
                self._conversation_len_before_first_response -= 1

            if not is_preexisting_self_response:
                self_message = self.apply_get_and_append_assistant_message()
                self_response = self_message.content

            # check if the response is valid:
            with self._app_temporarily_set_panel_status(PanelNames.FEEDBACK, 'Rule-based check ...'):
                self._app_send_prompt(PanelNames.FEEDBACK)
                response_error = None
                extracted_text = None
                try:
                    with self._is_extracting.temporary_set(True):
                        extracted_text = self._check_response_and_get_extracted_text(self_response)
                except SelfResponseError as e:
                    response_error = e
                if not response_error:
                    try:
                        self._check_extracted_text_and_update_valid_result(extracted_text)
                    except SelfResponseError as e:
                        response_error = e
                is_new_valid_result = self._valid_result_update_count > initial_valid_result_update_count

            if response_error:
                self._self_response_iteration_count -= response_error.add_iterations
                if response_error.bump_model:
                    try:
                        if response_error.bump_model == BumpModel.HIGHER_STRENGTH:
                            self.model_engine = self.model_engine.get_model_with_more_strength()
                        elif response_error.bump_model == BumpModel.HIGHER_CONTEXT:
                            self.model_engine = self.model_engine.get_model_with_more_context()
                        model_was_bumped = True
                    except ValueError:
                        model_was_bumped = False
                    if model_was_bumped:
                        msg = f"You seem totally drunk. Let's Bump you to {self.model_engine} and try again..."
                        print_and_log_red(msg)
            # else:
            #     self._app_send_prompt(PanelNames.FEEDBACK,
            #                           f'## {Symbols.CHECK_SYMBOL} Rule-based check passed.\n'
            #                           'The LLM response has successfully passed all rule-based checks.',
            #                           from_md=True,
            #                           sleep_for=PAUSE_AT_RULE_BASED_FEEDBACK)

            rewind = response_error.rewind if response_error else self.rewind_after_getting_a_valid_response

            # replace the response with a fresh looking response if needed:
            cycle_num = (len(self.conversation) - self._conversation_len_before_first_response + 1) // 2
            if (rewind == Rewind.AS_FRESH or rewind == Rewind.AS_FRESH_CORRECTION and cycle_num > 1) \
                    and extracted_text:
                self.apply_delete_messages(SingleMessageDesignation(-1))
                self.apply_append_surrogate_message(
                    self._convert_extracted_text_to_fresh_looking_response(extracted_text))
            # add the rule-based error message:
            if response_error:
                self.apply_append_user_message(self._convert_response_error_to_error_message(response_error),
                                               sleep_for=PAUSE_AT_RULE_BASED_FEEDBACK)

            # rewind:
            if rewind == Rewind.RESTART:
                self._rewind_conversation_to_first_response()
            elif rewind == Rewind.ACCUMULATE or rewind is None:
                pass
            elif rewind == Rewind.REGENERATE:
                assert response_error
                self.apply_delete_messages(RangeMessageDesignation.from_(-2, -1))
            elif rewind == Rewind.AS_FRESH_CORRECTION:
                assert response_error
                self._rewind_conversation_to_first_response(offset=2, last=-3)
            elif rewind == Rewind.DELETE_ALL:
                self._rewind_conversation_to_first_response(-1)  # delete including the user initiation prompt
            elif rewind == Rewind.AS_FRESH:
                self._rewind_conversation_to_first_response(offset=0, last=-3 if response_error else -2)

            if not response_error:
                assert is_new_valid_result
                return True, is_new_valid_result
        else:
            return False, is_new_valid_result

    def _get_valid_result(self) -> Union[Product, Any]:
        if not self._has_valid_result:
            raise FailedCreatingProductException('Failed to create a valid result.')
        return self.valid_result

    def _post_run(self):
        """
        Post-run actions.
        """
        pass

    def _run_and_return_termination_reason(self, *args, **kwargs) -> Any:
        """
        Run the conversation.
        """
        return self._iterate_until_valid_response()

    def run_and_get_valid_result_and_termination_reason(self, *args, **kwargs) -> Tuple[Tuple[bool, bool], Any]:
        """
        Run the conversation until we get a valid result.
        Return whether the conversation is converged and whether we have a new valid result.
        """
        self.initialize_conversation_if_needed()
        termination_reason = self._run_and_return_termination_reason(*args, **kwargs)
        result = self._get_valid_result()  # raises FailedCreatingProductException if no valid result
        self._post_run()
        return result, termination_reason

    def run_and_get_valid_result(self, *args, **kwargs) -> Any:
        """
        Run the conversation until we get a valid result.
        Return the valid result.
        """
        if not self._has_valid_result:
            self.run_and_get_valid_result_and_termination_reason(*args, **kwargs)
        return self._get_valid_result()



from __future__ import annotations

from abc import abstractmethod, ABC
from dataclasses import dataclass
from typing import Union, Optional, List

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from data_to_paper import Conversation


@dataclass(frozen=True)
class MessageDesignation(ABC):
    """
    a base class for indicating specific messages, or range of messages, within a conversation.
    """

    @abstractmethod
    def get_message_nums(self, conversation) -> List[int]:
        """
        Returns a list of indices indicating chosen messages in `conversation`.
        """
        pass

    @abstractmethod
    def __str__(self):
        pass


@dataclass(frozen=True)
class SingleMessageDesignation(MessageDesignation):
    """
    Indicates a single message by its tag or position.

    tag: indicates message tag (str), or position (int).
         negative values indicates counting from the end (-1 is the last message).
    """
    tag: Union[str, int]
    off_set: int = 0

    def get_message_num(self, conversation: Conversation) -> int:
        tag = self.tag if isinstance(self.tag, int) else conversation.get_message_index_by_tag(self.tag)
        if tag < 0:
            tag = len(conversation) + tag
        return tag + self.off_set

    def get_message_nums(self, conversation: Conversation) -> List[int]:
        return [self.get_message_num(conversation)]

    def __str__(self):
        if self.off_set:
            return f"<{self.tag} {self.off_set:+d}>"
        else:
            return f"<{self.tag}>"


@dataclass(frozen=True)
class RangeMessageDesignation(MessageDesignation):
    """
    Indicates a range of messages.

    start: first message, indicated, by tag (str), by index (int) or as SingleMessageDesignation.
    end: last message (including), indicated, by tag (str), by index (int) or as SingleMessageDesignation.
    """

    start: SingleMessageDesignation
    end: SingleMessageDesignation

    def get_message_nums(self, conversation: Conversation) -> List[int]:
        return list(range(self.start.get_message_num(conversation), self.end.get_message_num(conversation) + 1))

    def __str__(self):
        return f"{self.start} - {self.end}"

    @classmethod
    def from_(cls,
              start: Optional[Union[str, int, SingleMessageDesignation]] = None,
              end: Optional[Union[str, int, SingleMessageDesignation]] = None):
        start = 0 if start is None else start
        if not isinstance(start, SingleMessageDesignation):
            start = SingleMessageDesignation(start)
        end = -1 if end is None else end
        if not isinstance(end, SingleMessageDesignation):
            end = SingleMessageDesignation(end)
        return cls(start, end)


GeneralMessageDesignation = Optional[Union[MessageDesignation, str, int, List[Union[MessageDesignation, str, int]]]]


def convert_general_message_designation_to_list(designations: GeneralMessageDesignation
                                                ) -> List[Union[MessageDesignation, str, int]]:
    if designations is None:
        return []
    if isinstance(designations, list):
        return designations
    if isinstance(designations, tuple):
        return list(designations)
    return [designations]


def convert_general_message_designation_to_int_list(designations: GeneralMessageDesignation,
                                                    conversation: Conversation) -> List[int]:
    indices = set()
    for designation in convert_general_message_designation_to_list(designations):
        if isinstance(designation, tuple):
            designation = RangeMessageDesignation.from_(*designation)
        elif not isinstance(designation, MessageDesignation):
            designation = SingleMessageDesignation(designation)
        indices |= set(designation.get_message_nums(conversation))
    indices = sorted(indices)
    return indices



import re
from typing import Iterable

import regex

from .exceptions import UnwantedCommandsUsedInLatex

CHARS = {
    '&': r'\&',
    '%': r'\%',
    '#': r'\#',
    '_': r'\_',
    '$': r'\$',
    '~': r'\textasciitilde{}',
    '^': r'\textasciicircum{}',
    '<': r'$<$',
    '>': r'$>$',
    '≤': r'$\leq$',
    '≥': r'$\geq$',
    '≠': r'$\neq$',
    '±': r'$\pm$',
    '×': r'$\times$',
    '÷': r'$\div$',
    '°': r'$^{\circ}$',
    '∞': r'$\infty$',
    '√': r'$\sqrt{}$',
    '∑': r'$\sum$',
    '∏': r'$\prod$',
    '|': r'\textbar{}',
    '∈': r'$\in$',
    '∉': r'$\notin$',
    '∀': r'$\forall$',
    '∃': r'$\exists$',
    '∅': r'$\emptyset$',
    '\xe2': r'\^a',
    '\u200b': r'',  # zero width space
    '\u202f': r' ',  # narrow no-break space
}

NON_UTF8_CHARS = {
    '–': r'--',
    '’': r"",
    '≤': r'$\leq$',
    '≥': r'$\geq$',
    '±': r'+-',
    '²': r'$^2$',
    '³': r'$^3$',
    '¼': r'$\frac{1}{4}$',
    '½': r'$\frac{1}{2}$',
    '¾': r'$\frac{3}{4}$',
    '×': r'$\times$',
    '÷': r'$\div$',
    '°': r'$^{\circ}$',
    '∞': r'$\infty$',
    '√': r'$\sqrt{}$',
    '∑': r'$\sum$',
    '∏': r'$\prod$',
    '∈': r'$\in$',
    '∉': r'$\notin$',
    '∀': r'$\forall$',
    '∃': r'$\exists$',
    '∅': r'$\emptyset$',
    '∆': r'$\Delta$',
    '∇': r'$\nabla$',
    '∂': r'$\partial$',
    '“': r'"',
    '”': r'"',
}

assert all(len(c) == 1 for c in CHARS.keys())

MATH_PATTERN = r"""
(?<!\\)    # negative look-behind to make sure start is not escaped
(?:        # start non-capture group for all possible match starts
  # group 1, match dollar signs only
  # single or double dollar sign enforced by look-arounds
  ((?<!\$)\${1,2}(?!\$))|
  # group 2, match escaped parenthesis
  (\\\()|
  # group 3, match escaped bracket
  (\\\[)|
  # group 4,
  (\\begin\{(?:equation\*?|align\*?)\})|
  # group 5, match table and figure environments
  (\\begin\{(?:figure|lstlisting)\})|
  # group 6, match non-typesetting commands
  (\\(?:ref|label|autoref)\{)
)
# if group 1 was start
(?(1)
  # non greedy match everything in between
  # group 1 matches do not support recursion
  (.*?)(?<!\\)
  # match ending double or single dollar signs
  (?<!\$)\1(?!\$)|
# else
(?:
  # greedily and recursively match everything in between
  # groups 2, 3, 4, and 5 support recursion
  ((?:.|\n|\r)*?(?R)?(?:.|\n|\r)*?)(?<!\\)
  (?:
    # if group 2 was start, escaped parenthesis is end
    (?(2)\\\)|  
    # if group 3 was start, escaped bracket is end
    (?(3)\\\]|     
    # if group 4 was start, match end equation or end align
    (?(4)\\end\{(?:equation\*?|align\*?)\}| 
    # if group 5 was start, match end figure or end table
    (?(5)\\end\{(?:figure|lstlisting)\}|
    # else, match end of non-typesetting command
    \})
  )
)))))
"""

TABLES_CHARS = {
    r'>': r'$>$',
    r'<': r'$<$',
    r'=': r'$=$',
    r'|': r'\textbar{}',
}


def escape_special_chars_and_symbols_in_table(table: str,
                                              begin: str = r'\begin{tabular}',
                                              end: str = r'\end{tabular}') -> str:
    """
    Apply replace_special_chars to the tabular part of a latex table.
    """
    # extract the tabular part from the table using split
    if begin not in table:
        raise ValueError(f'The Table does not contain the begin command: {begin}')
    if end not in table:
        raise ValueError(f'The Table does not contain the end command: {end}')
    before_tabular, tabular_part = table.split(begin, 1)
    tabular_part, after_tabular = tabular_part.split(end, 1)
    tabular_part = process_latex_text_and_math(tabular_part, _process_table_part)
    return before_tabular + begin + tabular_part + end + after_tabular


def _process_table_part(tabular_part: str) -> str:
    pattern = re.compile(r'(%s)' % '|'.join(re.escape(key) for key in TABLES_CHARS.keys()))
    repl_func = lambda match: TABLES_CHARS[match.group(1)]
    return re.sub(pattern, repl_func, tabular_part)


def replace_special_latex_chars(text):
    chars = ''.join(CHARS.keys())
    pattern = fr'(?<!\\)([{chars}])'
    repl_func = lambda match: CHARS[match.group(1)]
    return re.sub(pattern, repl_func, text)


def replace_non_utf8_chars(text):
    chars = ''.join(NON_UTF8_CHARS.keys())
    pattern = fr'(?<!\\)([{chars}])'
    repl_func = lambda match: NON_UTF8_CHARS[match.group(1)]
    return re.sub(pattern, repl_func, text)


def process_inside_and_outside_command(latex, inside_func, outside_func):
    # Split the latex string into parts outside and within \caption{...}
    parts = re.split(pattern=r'(\\caption\{.*?\})', string=latex, flags=re.DOTALL)

    # Process each part using the appropriate function
    processed_parts = [outside_func(part) if not part.startswith(r'\caption') else r'\caption{' + inside_func(
        part[len(r'\caption{'):-1]) + '}' for part in parts]

    # Reassemble the parts
    processed_latex = ''.join(processed_parts)

    return processed_latex


def process_latex_text_and_math(text, process_text=replace_special_latex_chars, process_math=None):
    if process_math is None:
        process_math = lambda x: x

    result = []
    last_end = 0

    for match in regex.finditer(MATH_PATTERN, text, flags=regex.VERBOSE):
        non_math_part = text[last_end:match.start()]

        processed_part = process_text(non_math_part)
        result.append(processed_part)

        possibly_math_part = match.group()
        # find `\caption{...} parts in possibly_math_part and apply escaping on what's inside the curly braces
        # apply process_math on possibly_math_part without the `\caption{...}` and then add the processed caption

        processed_math_part = process_inside_and_outside_command(possibly_math_part, process_text, process_math)

        result.append(processed_math_part)

        last_end = match.end()

    # Process the remaining non-math part after the last match
    non_math_part = text[last_end:]
    processed_part = process_text(non_math_part)
    result.append(processed_part)

    return "".join(result)


def wrap_as_latex_code_output(paragraph):
    return "\\begin{codeoutput}\n" + paragraph + "\n\\end{codeoutput}"


def check_usage_of_un_allowed_commands(latex_content: str, unwanted_commands: Iterable[str]):
    unwanted_commands_used = [c for c in unwanted_commands if c in latex_content]
    if unwanted_commands_used:
        raise UnwantedCommandsUsedInLatex(unwanted_commands_used)



from typing import Iterable, Optional, Union, Set

import numpy as np

FIELDS_TO_NAMES = {
    'bibtex_id': 'ID',
    'title': 'Title',
    'journal': 'Journal',
    'journal_and_year': 'Journal and year',
    'tldr': 'TLDR',
    'abstract': 'Abstract',
    'year': 'Year',
    'influence': 'Citation influence',
    'query': 'Query',
    'search_rank': 'Search rank',
    'embedding_similarity': 'Embedding similarity',
}


class Citation(dict):
    """
    A citation of a paper.
    """

    def __init__(self, *args, search_rank: int = None, query: Union[str, Set[str]] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.search_rank = search_rank
        self.query = query

    def __key(self):
        return self.bibtex_id

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    @property
    def bibtex(self) -> str:
        raise NotImplementedError

    @property
    def bibtex_id(self) -> str:
        raise NotImplementedError

    @property
    def title(self) -> Optional[str]:
        return None

    @property
    def tldr(self) -> Optional[str]:
        return None

    @property
    def abstract(self) -> Optional[str]:
        return None

    @property
    def journal(self) -> Optional[str]:
        return None

    @property
    def year(self) -> Optional[str]:
        return None

    @property
    def influence(self) -> int:
        return 0

    @property
    def embedding(self) -> Optional[np.ndarray]:
        return None

    @property
    def journal_and_year(self) -> Optional[str]:
        if self.journal is None or self.year is None:
            return None
        return f'{self.journal} ({self.year})'

    def get_embedding_similarity(self, embedding_target: Optional[np.ndarray]) -> float:
        if self.embedding is None or embedding_target is None:
            return 0
        return np.dot(self.embedding, embedding_target) / \
            (np.linalg.norm(self.embedding) * np.linalg.norm(embedding_target))

    def pretty_repr(self,
                    fields: Iterable[str] = ('bibtex_id', 'title', 'journal_and_year', 'tldr', 'influence'),
                    is_html: bool = False,
                    embedding_target: float = None,
                    ) -> str:
        """
        Get a pretty representation of the citation.
        Allows specifying which fields to include.
        """
        s = ''
        for field in fields:
            name = FIELDS_TO_NAMES[field]
            if field == 'embedding_similarity':
                value = self.get_embedding_similarity(embedding_target)
                value = None if value is None else round(value, 2)
            else:
                value = getattr(self, field, field)
            if field == 'query' and isinstance(value, set):
                value = sorted(value)
            if value is None:
                continue
            if is_html:
                # name in bold
                s += f'<b>{name}</b>: {value}<br>'
            else:
                s += f'{name}: {repr(value)}\n'
        if is_html:
            s += '<br>'
        else:
            s += '\n'
        return s

    def __str__(self):
        return self.pretty_repr()

    def __repr__(self):
        return self.__str__()



import pickle

from data_to_paper.env import FOLDER_FOR_RUN
from data_to_paper.research_types.hypothesis_testing.check_df_to_funcs.df_checker import \
    check_analysis_df, check_displayitem_df
from data_to_paper.run_gpt_code.overrides.dataframes.df_with_attrs import InfoDataFrameWithSaveObjFuncCall

files_starting_with_df = list(FOLDER_FOR_RUN.glob('df*'))

if __name__ == '__main__':
    for filepath in files_starting_with_df:
        with open(filepath, 'rb') as file:
            df = pickle.load(file)
            assert isinstance(df, InfoDataFrameWithSaveObjFuncCall)
        if df.get_prior_filename() is None:
            issues = check_analysis_df(df, output_folder=FOLDER_FOR_RUN)
        else:
            issues = check_displayitem_df(df, output_folder=FOLDER_FOR_RUN)

        print(f'Issues with {filepath}:')
        for issue in issues:
            print(issue)



from dataclasses import dataclass
from typing import Tuple, Optional, Dict, Type

from data_to_paper.base_products.file_descriptions import DataFileDescriptions, DataFileDescription
from data_to_paper.base_steps import LatexReviewBackgroundProductsConverser, BackgroundProductsConverser, \
    ReviewBackgroundProductsConverser, PythonDictWithDefinedKeysReviewBackgroundProductsConverser
from data_to_paper.base_steps.base_products_conversers import ProductsConverser
from data_to_paper.base_steps.result_converser import Rewind
from data_to_paper.code_and_output_files.code_and_output import CodeAndOutput
from data_to_paper.latex import extract_latex_section_from_response
from data_to_paper.latex.latex_doc import LatexDocument

from data_to_paper.research_types.hypothesis_testing.coding.base_code_conversers import \
    BaseScientificCodeProductsHandler, BaseScientificCodeProductsGPT
from data_to_paper.text import dedent_triple_quote_str
from data_to_paper.utils.nice_list import NiceDict
from data_to_paper.utils.replacer import Replacer
from data_to_paper.utils.types import ListBasedSet


@dataclass
class BaseScientificPostCodeProductsHandler(BaseScientificCodeProductsHandler):
    background_product_fields: Tuple[str, ...] = None
    goal_noun: str = '{code_name} code'
    goal_verb: str = 'write'
    max_reviewing_rounds = 0

    def __post_init__(self):
        if self.background_product_fields is None:
            self.background_product_fields = ('all_file_descriptions', f'codes_and_outputs:{self.code_step}',)
        super().__post_init__()

    @property
    def code_and_output(self):
        return self.products.codes_and_outputs[self.code_step]

    @property
    def created_output_filenames(self):
        return self.code_and_output.created_files.get_all_created_files()


@dataclass
class RequestCodeExplanation(BaseScientificPostCodeProductsHandler, LatexReviewBackgroundProductsConverser):
    goal_noun: str = 'explanation of the {code_name} code'
    background_product_fields: Tuple[str, ...] = ('data_file_descriptions',)
    max_reviewing_rounds: int = 0
    rewind_after_end_of_review: Rewind = Rewind.DELETE_ALL
    rewind_after_getting_a_valid_response: Optional[Rewind] = Rewind.ACCUMULATE
    should_remove_citations_from_section: bool = True
    section_names: Tuple[str, ...] = ('Code Explanation',)

    def __post_init__(self):
        self.background_product_fields = self.background_product_fields + ('codes:' + self.code_step,)
        BaseScientificPostCodeProductsHandler.__post_init__(self)
        LatexReviewBackgroundProductsConverser.__post_init__(self)

    mission_prompt: str = dedent_triple_quote_str("""
        Please return a triple-backtick Latex Block explaining what the code above does. 
        Do not provide a line-by-line explanation, rather provide a \t
        high-level explanation of the code in a language suitable for a Methods section of a research \t
        paper.
        Structure the explanation according to the steps of the analysis, and explain the purpose of each step, \t
        and how it was implemented in the code.
        There is no need to explain trivial parts, like reading/writing a file, etc.  

        Your explanation should be written in LaTeX, and should be enclosed within a LaTeX Code Block, like this:

        ```latex
        \\section{Code Explanation}
        <your code explanation here>
        ```

        Remember to enclose your explanation within a LaTeX Code Block, so that I can easily copy-paste it!
        """)

    request_triple_quote_block: Optional[str] = dedent_triple_quote_str("""
        Your code explanation should be enclosed within a triple-backtick "latex" block.
        """)

    requesting_output_explanation: str = dedent_triple_quote_str("""
        Also explain what does the code write into the files(s): {created_output_filenames}.    
        """)

    def run_and_get_valid_result(self):
        result = super().run_and_get_valid_result()
        return extract_latex_section_from_response(result[0], 'Code Explanation', keep_tags=False)


@dataclass
class ExplainCreatedDataframe(BaseScientificPostCodeProductsHandler, BackgroundProductsConverser):
    goal_noun: str = 'explanation of the files created by the {code_name} code'

    def __post_init__(self):
        self.background_product_fields = self.background_product_fields + ('codes:' + self.code_step,)
        BaseScientificPostCodeProductsHandler.__post_init__(self)
        BackgroundProductsConverser.__post_init__(self)

    mission_prompt: str = None
    background_product_fields: Tuple[str, ...] = ('all_file_descriptions', 'research_goal')
    requesting_explanation_for_a_new_dataframe: str = dedent_triple_quote_str("""
        The code creates a new file named "{dataframe_file_name}", with the following columns: 
        {columns}.

        Explain the content of the file, and how it was derived from the original data. 
        Importantly: do NOT explain the content of columns that are already explained for the \t
        original dataset (see above DESCRIPTION OF THE DATASET).
        """)

    requesting_explanation_for_a_modified_dataframe: str = dedent_triple_quote_str("""
        Explain the content of all the new or modified columns of "{dataframe_file_name}".

        Return your explanation as a dictionary, where the keys are the column names {columns}, \t
        and the values are the strings that explain the content of each column.

        All information you think is important should be encoded in this dictionary. 
        Do not send additional free text beside the text in the dictionary.  
        """)

    def ask_for_created_files_descriptions(self) -> Optional[DataFileDescriptions]:
        self.initialize_conversation_if_needed()
        data_folder = self.products.data_file_descriptions.data_folder
        dataframe_operations = self.code_and_output.dataframe_operations
        data_file_descriptions = DataFileDescriptions(data_folder=data_folder)
        saved_ids_filenames = dataframe_operations.get_saved_ids_filenames()
        # sort the saved ids by their filename, so that the order of the questions will be consistent between runs:
        saved_ids_filenames = sorted(saved_ids_filenames, key=lambda saved_id_filename: saved_id_filename[1])

        for saved_df_id, saved_df_filename in saved_ids_filenames:
            read_filename = dataframe_operations.get_read_filename(saved_df_id)
            saved_columns = ListBasedSet(dataframe_operations.get_save_columns(saved_df_id))
            creation_columns = ListBasedSet(dataframe_operations.get_creation_columns(saved_df_id))
            changed_columns = ListBasedSet(dataframe_operations.get_changed_columns(saved_df_id))
            added_columns = saved_columns - creation_columns
            if read_filename is None:
                # this saved dataframe was created by the code, not read from a file
                columns = saved_columns
                response = ReviewBackgroundProductsConverser.from_(
                    self,
                    is_new_conversation=False,
                    max_reviewing_rounds=0,
                    rewind_after_end_of_review=Rewind.DELETE_ALL,
                    rewind_after_getting_a_valid_response=Rewind.ACCUMULATE,
                    goal_noun='the content of the dataframe',
                    mission_prompt=Replacer(self, self.requesting_explanation_for_a_new_dataframe,
                                            kwargs={'dataframe_file_name': saved_df_filename,
                                                    'columns': list(columns)}),
                ).run_and_get_valid_result()
                description = f'This csv file was created by the {self.code_name} code.\n' \
                              f'{response}\n'
                data_file_description = DataFileDescription(file_path=saved_df_filename, description=description,
                                                            originated_from=None)
            else:
                # this saved dataframe was read from a file
                columns = added_columns | changed_columns
                columns_to_explanations = PythonDictWithDefinedKeysReviewBackgroundProductsConverser.from_(
                    self,
                    is_new_conversation=False,
                    max_reviewing_rounds=0,
                    rewind_after_end_of_review=Rewind.DELETE_ALL,
                    rewind_after_getting_a_valid_response=Rewind.ACCUMULATE,
                    requested_keys=columns,
                    goal_noun='dictionary that explains the columns of the dataframe',
                    mission_prompt=Replacer(self,
                                            self.requesting_explanation_for_a_modified_dataframe,
                                            kwargs={'dataframe_file_name': saved_df_filename,
                                                    'columns': list(columns)}),
                    value_type=Dict[str, str],
                ).run_and_get_valid_result()

                new_columns_to_explanations = \
                    {column: explanation for column, explanation in columns_to_explanations.items()
                     if column in added_columns}
                modified_columns_to_explanations = \
                    {column: explanation for column, explanation in columns_to_explanations.items()
                        if column not in added_columns}

                if len(modified_columns_to_explanations) > 0:
                    modified_columns_str = f'\nWe modified these columns:\n' \
                                           f'{NiceDict(modified_columns_to_explanations)}\n'
                else:
                    modified_columns_str = ''

                if len(new_columns_to_explanations) > 0:
                    new_columns_str = f'\nWe added these columns:\n' \
                                      f'{NiceDict(new_columns_to_explanations)}\n'
                else:
                    new_columns_str = ''

                description = f'This csv file was created by our {self.code_name} code ' \
                              f'from the file "{read_filename}".\n' \
                              f'{modified_columns_str}' \
                              f'{new_columns_str}'
                data_file_description = DataFileDescription(file_path=saved_df_filename, description=description,
                                                            originated_from=read_filename)

            data_file_descriptions.append(data_file_description)

        return data_file_descriptions


@dataclass
class RequestCodeProducts(BaseScientificCodeProductsHandler, ProductsConverser):
    code_writing_class: Type[BaseScientificCodeProductsGPT] = None
    explain_code_class: Optional[Type[RequestCodeExplanation]] = RequestCodeExplanation
    explain_created_files_class: Optional[Type[ExplainCreatedDataframe]] = None
    latex_document: LatexDocument = None

    def _save_code_to_file(self, code_step: str, code_and_output: CodeAndOutput):
        """
        Save the code to a file, only if self.output_directory was defined.
        """
        if self.output_directory is None:
            return
        with open(f'{self.output_directory}/{code_step}.py', 'w', encoding='utf-8') as f:
            f.write(code_and_output.code)

    def get_code_and_output(self) -> CodeAndOutput:
        code_writing = self.code_writing_class.from_(self)
        assert code_writing.code_step == self.code_step
        return code_writing.get_code_and_output()

    def _get_description_of_created_files(self) -> Optional[DataFileDescriptions]:
        return self.explain_created_files_class.from_(
            self,
            is_new_conversation=None,
            conversation_name=self.conversation_name + ' - Created Files',
            code_step=self.code_step,
        ).ask_for_created_files_descriptions()

    def _get_code_explanation(self) -> str:
        return self.explain_code_class.from_(
            self,
            is_new_conversation=None,
            conversation_name=self.conversation_name + ' - Explanation',
            code_step=self.code_step,
        ).run_and_get_valid_result()

    def get_code_and_output_and_descriptions(self) -> CodeAndOutput:
        code_and_output = self.get_code_and_output()
        self.products.codes_and_outputs[self.code_step] = code_and_output
        if self.explain_code_class:
            code_and_output.code_explanation = self._get_code_explanation()
        if self.explain_created_files_class and code_and_output.created_files.get_created_data_files():
            code_and_output.description_of_created_files = self._get_description_of_created_files()
        self._save_code_to_file(self.code_step, code_and_output)
        return code_and_output



from __future__ import annotations
import functools
from typing import Iterable
from dataclasses import dataclass

from data_to_paper.env import TRACK_P_VALUES
from ...attr_replacers import SystematicMethodReplacerContext, SystematicFuncReplacerContext, AttrReplacer
from ...run_issues import CodeProblem, RunIssue
from ..pvalue import convert_to_p_value, PValue, TrackPValueCreationFuncs

MULTITEST_FUNCS_AND_PVAL_INDEXES = [
    ('multipletests', 1),
    ('fdrcorrection_twostage', 1),
    ('fdrcorrection', 1),
]


ANOVA_FUNCS = [
    'anova_single',
    'anova1_lm_single',
    'anova2_lm_single',
    'anova3_lm_single',
    'anova_lm',
]


def _set_result_attr_or_modify_cache(result, attr, pvalues):
    try:
        setattr(result, attr, pvalues)
    except AttributeError:
        if attr in getattr(result, '_cache', {}):
            result._cache[attr] = pvalues


def custom_summary2_method(*args, **kwargs):
    raise RunIssue.from_current_tb(
        category="Statsmodels: good practices",
        issue=f"Do not use any methods of the object created by summary2().",
        instructions=f"Use instead the `tables` attribute of the object returned by summary2(), "
                     f"from which you can extract the numerical values that you need.\n"
                     f"Do not transform numerical values to strings.",
        code_problem=CodeProblem.RuntimeError,
    )


@dataclass
class StatsmodelsFitPValueOverride(SystematicMethodReplacerContext, TrackPValueCreationFuncs):
    """
    A context manager that replaces the pvalues attribute of all fit functions in statsmodels with a
    PValue.
    """
    package_names: Iterable[str] = ('statsmodels', )

    def _get_all_modules(self) -> list:
        from statsmodels.regression import linear_model
        from statsmodels.genmod import generalized_linear_model
        from statsmodels.discrete import discrete_model
        return [linear_model, generalized_linear_model, discrete_model]

    def _should_replace(self, parent, attr_name, attr) -> bool:
        return attr_name.startswith('fit')

    def _get_summary2_func(self, obj, results, created_by: str):
        """
        Get the overridden summary2 function of a statsmodels class.
        """
        original_summary2 = results.summary2

        def custom_summary2(*args, **kwargs):
            """
            A custom summary2 function that replaces the pvalues attribute of the summary tables with a PValue objects.
            Replaces "P>|t|" and "P>|z|" with PValue objects.
            """
            with PValue.BEHAVE_NORMALLY.temporary_set(True):
                result = original_summary2(*args, **kwargs)

            tables = result.tables
            table1 = tables[1]

            for columns in table1.columns:
                if columns.startswith('P>'):
                    table1[columns] = convert_to_p_value(table1[columns], created_by=created_by,
                                                         var_name=columns, context=self)
                else:
                    # We need to convert the columns to float because the summary2 function
                    # converts all the columns to object type even if there is just one PValue column.
                    try:
                        table1[columns] = table1[columns].astype(float)
                    except ValueError:
                        pass

            unallowed_method = ['as_text', 'as_latex', 'as_html', 'as_csv']
            for method_name in unallowed_method:
                if hasattr(result, method_name):
                    setattr(result, method_name, custom_summary2_method)

            return result

        return custom_summary2

    def _get_summary_func(self, original_func):
        """
        Get the overridden summary function of a statsmodels class.
        """

        def custom_summary(*args, **kwargs):
            """
            Prevents the use of the summary function.
            """
            raise RunIssue.from_current_tb(
                category="Statsmodels: good practices",
                issue=f"Do not use the `summary` function of statsmodels.",
                instructions=f"Use the `summary2` function instead, "
                             f"from which you can extract the numerical values that you need.\n"
                             f"Do not transform numerical values to strings.",
                code_problem=CodeProblem.RuntimeError,
            )

        return custom_summary

    def _get_custom_wrapper(self, parent, attr_name, original_func):

        @functools.wraps(original_func)
        def wrapped(obj, *args, **kwargs):
            result = original_func(obj, *args, **kwargs)
            if self._is_called_from_user_script():
                if hasattr(obj, '_prior_fit_results') and obj._prior_fit_results is result:
                    #  TODO: Can likely be removed.
                    # This behavior happen on sklean.
                    # But not on statsmodels. So, we don't need to handle this case.
                    # See test_statsmodels_multiple_fit_calls_normal_has_no_bug
                    raise RunIssue.from_current_tb(
                        category='Statsmodels: good practices',
                        issue=f"The `{original_func.__name__}` function was already called on this object. ",
                        instructions=f"Multiple calls should be avoided as the same result instance is returned again.",
                        code_problem=CodeProblem.RuntimeError,
                    )
                obj._prior_fit_results = result

            if TRACK_P_VALUES:
                # Replace the pvalues attribute if it exists
                created_by = obj.__class__.__name__
                pvalue_detected = False
                for attr in ['pvalues', 'f_pvalue', 'pvalue']:
                    if not hasattr(result, attr):
                        continue
                    pvalues = getattr(result, attr)
                    pvalues = convert_to_p_value(pvalues, created_by=created_by, context=self,
                                                 raise_on_nan=attr != 'f_pvalue')
                    pvalue_detected = True
                    _set_result_attr_or_modify_cache(result, attr, pvalues)

                if hasattr(result, 'summary2'):
                    result.summary2 = self._get_summary2_func(obj, result, created_by)
                    pvalue_detected = True

                if hasattr(result, 'summary'):
                    original_summary = result.summary
                    result.summary = self._get_summary_func(original_summary)
                    pvalue_detected = True

                if hasattr(result, 'eigenvals'):
                    eigenvals = result.eigenvals
                    if eigenvals is not None:
                        min_eigenval = eigenvals[-1]
                        assert min_eigenval == min(eigenvals)
                        if min_eigenval < 1e-10:
                            self.issues.append(RunIssue.from_current_tb(
                                category='Statsmodels: eigenvalues too small',
                                issue="The eigenvalues of the covariance matrix are too small. "
                                      "This might indicate that there are strong multicollinearity problems "
                                      "or that the design matrix is singular.",
                                instructions="Try to remove redundant features.",
                                code_problem=CodeProblem.RuntimeError,
                            ))

                if pvalue_detected:
                    self._add_pvalue_creating_func(created_by)
            return result

        return wrapped


@dataclass
class StatsmodelsMultitestPValueOverride(SystematicFuncReplacerContext, TrackPValueCreationFuncs):
    package_names: Iterable[str] = ('statsmodels', )
    obj_import_str: str = 'statsmodels.stats.multitest'

    def _should_replace(self, module, func_name, func) -> bool:
        return func_name in [func_name for func_name, _ in MULTITEST_FUNCS_AND_PVAL_INDEXES]

    def _get_custom_wrapper(self, parent, attr_name, original_func):

        @functools.wraps(original_func)
        def wrapped(*args, **kwargs):
            result = original_func(*args, **kwargs)

            if TRACK_P_VALUES:
                # Replace the pvalues attribute if it exists
                try:
                    func_name, pval_index = [x for x in MULTITEST_FUNCS_AND_PVAL_INDEXES if x[0] == attr_name][0]
                    result = list(result)
                    result[pval_index] = convert_to_p_value(result[pval_index], created_by=func_name, context=self)
                    result = tuple(result)
                    self._add_pvalue_creating_func(func_name)
                except (AttributeError, TypeError, ValueError):
                    pass
            return result

        return wrapped


@dataclass
class StatsmodelsAnovaPValueOverride(SystematicFuncReplacerContext, TrackPValueCreationFuncs):
    package_names: Iterable[str] = ('statsmodels', )
    obj_import_str: str = 'statsmodels.stats.anova'

    def _should_replace(self, module, func_name, func) -> bool:
        return func_name in ANOVA_FUNCS

    def _get_custom_wrapper(self, parent, attr_name, original_func):

        @functools.wraps(original_func)
        def wrapped(*args, **kwargs):
            result = original_func(*args, **kwargs)

            if TRACK_P_VALUES:
                # Replace the 'PR(>F)' column with PValue objects
                try:
                    for row_label in result.index:
                        result.loc[row_label, 'PR(>F)'] = convert_to_p_value(result.loc[row_label, 'PR(>F)'],
                                                                             created_by=attr_name,
                                                                             var_name=row_label,
                                                                             raise_on_nan=row_label != 'Residual',
                                                                             context=self)
                    self._add_pvalue_creating_func(attr_name)
                except (AttributeError, TypeError, ValueError):
                    pass
            return result

        return wrapped


@dataclass
class StatsmodelsMulticompPValueOverride(AttrReplacer, TrackPValueCreationFuncs):
    package_names: Iterable[str] = ('statsmodels', )
    obj_import_str: str = 'statsmodels.sandbox.stats.multicomp.TukeyHSDResults'
    attr: str = '__init__'

    def _get_wrapped_wrapper(self):

        def TukeyHSDResults__init__(obj, mc_object, results_table, q_crit, reject=None,
                                    meandiffs=None, std_pairs=None, confint=None, df_total=None,
                                    reject2=None, variance=None, pvalues=None):
            """
            A custom __init__ function for TukeyHSDResults that replaces the pvalues attribute with a PValue object.
            """
            pvalues = convert_to_p_value(pvalues, created_by='TukeyHSDResults')

            p_adj_index = 3
            for row in results_table[1:]:
                row[p_adj_index].data = convert_to_p_value(row[p_adj_index].data, created_by='TukeyHSDResults',
                                                           var_name=row[0].data, context=self)

            self._add_pvalue_creating_func('TukeyHSDResults')
            return self._original(obj, mc_object, results_table, q_crit, reject=reject,
                                  meandiffs=meandiffs, std_pairs=std_pairs, confint=confint, df_total=df_total,
                                  reject2=reject2, variance=variance, pvalues=pvalues)

        return TukeyHSDResults__init__



import collections
from enum import Enum

from typing import Generic, TypeVar, Iterable


class IndexOrderedEnum(Enum):

    def get_index(self):
        """
        Get the index of this enum value in the list of enum values.
        """
        return self._member_names_.index(self.name)

    def get_next(self):
        """
        Get the next enum value in the list.
        If this is the last value, a ValueError is raised.
        """
        try:
            return list(self.__class__)[(self.get_index() + 1)]
        except IndexError:
            raise ValueError(f"No next value after {self}")

    @classmethod
    def get_first(cls):
        return list(cls)[0]

    @classmethod
    def get_last(cls):
        return list(cls)[-1]

    def __eq__(self, other):
        if isinstance(other, IndexOrderedEnum):
            return self.get_index() == other.get_index()
        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, IndexOrderedEnum):
            return self.get_index() < other.get_index()
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, IndexOrderedEnum):
            return self.get_index() <= other.get_index()
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, IndexOrderedEnum):
            return self.get_index() > other.get_index()
        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, IndexOrderedEnum):
            return self.get_index() >= other.get_index()
        return NotImplemented

    def __hash__(self):
        return hash(self.get_index())


T = TypeVar('T')


class ListBasedSet(collections.abc.Set, Generic[T]):
    """
    Alternate set implementation favoring space over speed and stable ordering.
    also and not requiring the set elements to be hashable.
    """

    def __init__(self, iterable: Iterable = None):
        self.elements = lst = []
        if iterable is not None:
            for value in iterable:
                if value not in lst:
                    lst.append(value)

    def __iter__(self):
        return iter(self.elements)

    def __contains__(self, value):
        return value in self.elements

    def __len__(self):
        return len(self.elements)

    def __str__(self):
        # make it look like a set:
        return '{' + ', '.join(repr(e) for e in self) + '}'

    def __repr__(self):
        return f'{self.__class__.__name__}({self.elements})'

    def add(self, value):
        if value not in self.elements:
            self.elements.append(value)

    def remove(self, value):
        self.elements.remove(value)

    def update(self, other):
        for value in other:
            self.add(value)

    def union(self, other):
        return self.__class__(self.elements + list(other))


K = TypeVar('K')
V = TypeVar('V')


class MemoryDict(Generic[K, V]):
    def __init__(self):
        self._data = {}

    def __getitem__(self, key):
        return self._data[key][-1]['value']  # Retrieve the most recent value

    def __setitem__(self, key, value):
        if key not in self._data:
            self._data[key] = [{'name': 'default', 'value': value}]
        else:
            self._data[key].append({'name': 'default', 'value': value})

    def add_named_value(self, key, name, value):
        if key not in self._data:
            self._data[key] = [{'name': name, 'value': value}]
        else:
            self._data[key].append({'name': name, 'value': value})

    def get_named_value(self, key, name):
        if key in self._data:
            for item in reversed(self._data[key]):
                if item['name'] == name:
                    return item['value']
        return None

    def get_all_values(self, key):
        return [item['value'] for item in self._data.get(key, [])]

    def get_all_named_values(self, key):
        return [(item['name'], item['value']) for item in self._data.get(key, [])]

    def as_dict(self):
        return {key: self[key] for key in self._data}

    def items(self):
        return self.as_dict().items()

    def __delitem__(self, key):
        if key in self._data:
            del self._data[key]

    def __contains__(self, key):
        return key in self._data

    def __len__(self):
        return len(self._data)



from __future__ import annotations
from typing import Dict, Optional, Union

from pathlib import Path

from data_to_paper.conversation.stage import Stage
from data_to_paper.servers.json_dump import dump_to_json, load_from_json


class StageToCost(Dict[Optional[Stage], float]):
    """
    A dictionary that stores the api cost for each stage
    key of None is for the total costs of stages that were reset and deleted
    """

    def get_total_cost(self, with_deleted: bool = True) -> float:
        return sum(cost for stage, cost in self.items() if with_deleted or stage is not None)

    def delete_from_stage(self, stage: Stage):
        """
        delete_all_stages_following_stage
        store deleted stage costs in None
        """
        for key in list(self.keys()):
            if key is not None and key >= stage:
                cost = self.pop(key)
                self[None] = self.get(None, 0) + cost

    def save_to_json(self, path: Union[str, Path]):
        """
        Save the api cost to a json file
        Deleted stages are not saved
        """
        as_dict = {stage.value: round(cost, 2) for stage, cost in self.items() if stage is not None}
        dump_to_json(as_dict, path)

    @classmethod
    def load_from_json(cls, path: str) -> StageToCost:
        return cls(load_from_json(path))

    def as_html(self) -> str:
        if not self:
            return '<span style="color: white;">No API usage cost</span>'
        s = '<h2>API usage cost</h2>\n'
        s += '<p>\n'
        s += '<table style="color:white;">\n'
        for stage, cost in self.items():
            if stage is not None:
                s += f'<tr>\n<td>{stage.value}</td>\n<td>${cost:.2f}</td>\n</tr>\n'
        if None in self:
            s += '<tr>\n<td></td>\n<td></td>\n</tr>\n'
            s += f'<tr>\n<td>Deleted stages</td>\n<td>${self[None]:.2f}</td>\n</tr>\n'
        s += '</table>\n'
        s += '</p>\n'
        s += f'<h3>Total cost: ${self.get_total_cost():.2f}</h3>\n'
        return s



import os
import subprocess


SYSTEM_KWARGS = {}
if os.name == 'nt':
    SYSTEM_KWARGS = {'creationflags': subprocess.CREATE_NO_WINDOW}


def get_subprocess_kwargs(capture: bool = True):
    if capture:
        return {'check': True, 'stdout': subprocess.PIPE, 'stderr': subprocess.PIPE, **SYSTEM_KWARGS}
    return {'check': True, 'stdout': subprocess.DEVNULL, 'stderr': subprocess.DEVNULL, **SYSTEM_KWARGS}
